from pathlib import Path

import pytest
from transformers import EncoderDecoderConfig
from transformers.models.gpt2 import GPT2Config

import realtabformer
from realtabformer.realtabformer import ModelType, REaLTabFormer

RANDOM_SEED = 1029


def test_ModelType():
    assert ModelType.types() == ["tabular", "relational"]


def test_default_init():
    model_types = [ModelType.tabular, ModelType.relational]

    for model_type in model_types:
        rtf_model = REaLTabFormer(model_type)

        # Track the variables that we have tested to
        # make sure that all variables that will be
        # added or removed in the future will be caught
        # by the test.
        model_vars_tested = set()

        assert rtf_model.model_type == model_type
        model_vars_tested.add("model_type")

        # Check default dir arguments
        assert isinstance(rtf_model.checkpoints_dir, Path)
        assert rtf_model.checkpoints_dir.name == "rtf_checkpoints"
        model_vars_tested.add("checkpoints_dir")

        assert isinstance(rtf_model.samples_save_dir, Path)
        assert rtf_model.samples_save_dir.name == "rtf_samples"
        model_vars_tested.add("samples_save_dir")

        assert rtf_model.epochs == 100
        model_vars_tested.add("epochs")

        assert rtf_model.batch_size == 8
        model_vars_tested.add("batch_size")

        assert rtf_model.random_state == 1029
        model_vars_tested.add("random_state")

        assert rtf_model.train_size == 1
        model_vars_tested.add("train_size")

        assert rtf_model.early_stopping_patience == 5
        model_vars_tested.add("early_stopping_patience")

        assert rtf_model.early_stopping_threshold == 0
        model_vars_tested.add("early_stopping_threshold")

        assert rtf_model.mask_rate == 0
        model_vars_tested.add("mask_rate")

        assert rtf_model.numeric_nparts == 1
        model_vars_tested.add("numeric_nparts")

        assert rtf_model.numeric_precision == 4
        model_vars_tested.add("numeric_precision")

        assert rtf_model.numeric_max_len == 10
        model_vars_tested.add("numeric_max_len")

        if model_type == ModelType.tabular:
            with pytest.raises(AttributeError):
                # The argument `output_max_length` is not set
                # for the tabular model.
                assert rtf_model.output_max_length

            # Implicitly tests `_init_tabular`
            assert isinstance(rtf_model.tabular_config, GPT2Config)
            assert rtf_model.tabular_config.n_layer == 6
            model_vars_tested.add("tabular_config")
        else:
            assert rtf_model.output_max_length == 512
            model_vars_tested.add("output_max_length")

            assert rtf_model.freeze_parent_model
            model_vars_tested.add("freeze_parent_model")

            # Relational model
            assert rtf_model.parent_vocab is None
            model_vars_tested.add("parent_vocab")

            assert rtf_model.parent_gpt2_config is None
            model_vars_tested.add("parent_gpt2_config")

            assert rtf_model.parent_gpt2_state_dict is None
            model_vars_tested.add("parent_gpt2_state_dict")

            assert rtf_model.parent_col_transform_data is None
            model_vars_tested.add("parent_col_transform_data")

            # Implicitly tests `_init_relational`
            assert isinstance(rtf_model.relational_config, EncoderDecoderConfig)
            assert isinstance(rtf_model.relational_config.encoder, GPT2Config)
            assert isinstance(rtf_model.relational_config.decoder, GPT2Config)
            assert rtf_model.relational_config.encoder.n_layer == 6
            assert rtf_model.relational_config.decoder.n_layer == 6

            model_vars_tested.add("relational_config")

        # Validate the implicit default values in `training_args_kwargs`
        assert rtf_model.training_args_kwargs["evaluation_strategy"] == "steps"
        assert (
            rtf_model.training_args_kwargs["output_dir"]
            == rtf_model.checkpoints_dir.as_posix()
        )
        assert rtf_model.training_args_kwargs["evaluation_strategy"] == "steps"

        assert rtf_model.training_args_kwargs["metric_for_best_model"] == "loss"
        assert rtf_model.training_args_kwargs["overwrite_output_dir"] is True
        assert rtf_model.training_args_kwargs["num_train_epochs"] == rtf_model.epochs
        assert (
            rtf_model.training_args_kwargs["per_device_train_batch_size"]
            == rtf_model.batch_size
        )
        assert (
            rtf_model.training_args_kwargs["per_device_eval_batch_size"]
            == rtf_model.batch_size
        )

        assert rtf_model.training_args_kwargs["gradient_accumulation_steps"] == 4
        assert rtf_model.training_args_kwargs["remove_unused_columns"] is True
        assert rtf_model.training_args_kwargs["logging_steps"] == 100
        assert rtf_model.training_args_kwargs["save_steps"] == 100
        assert rtf_model.training_args_kwargs["eval_steps"] == 100
        assert rtf_model.training_args_kwargs["load_best_model_at_end"] is True
        assert (
            rtf_model.training_args_kwargs["save_total_limit"]
            == rtf_model.early_stopping_patience + 1
        )
        model_vars_tested.add("training_args_kwargs")

        # Validate empty-initialized attributes
        list_defaults = [
            "columns",
            "drop_na_cols",
            "processed_columns",
            "numeric_columns",
            "datetime_columns",
        ]
        for ld in list_defaults:
            assert (
                isinstance(getattr(rtf_model, ld), list)
                and len(getattr(rtf_model, ld)) == 0
            )

        dict_defaults = ["column_dtypes", "column_has_missing", "vocab", "col_idx_ids"]
        for dd in dict_defaults:
            assert (
                isinstance(getattr(rtf_model, dd), dict)
                and len(getattr(rtf_model, dd)) == 0
            )

        none_defaults = [
            "model",
            "tabular_max_length",
            "relational_max_length",
            "tabular_col_size",
            "relational_col_size",
            "experiment_id",
            "col_transform_data",
            "in_col_transform_data",
            "target_col",
            "trainer_state",
        ]
        for nd in none_defaults:
            assert getattr(rtf_model, nd) is None

        assert rtf_model.realtabformer_version == realtabformer.__version__
        model_vars_tested.add("realtabformer_version")

        model_vars_tested.update(list_defaults)
        model_vars_tested.update(dict_defaults)
        model_vars_tested.update(none_defaults)

        model_vars = set(vars(rtf_model))

        print(model_vars.difference(model_vars_tested))

        assert len(model_vars.difference(model_vars_tested)) == 0
        assert len(model_vars) == len(model_vars_tested), f"{model_vars}...{model_vars_tested}"


# def test_tabular_init():
#     training_args_kwargs = dict(
#         logging_steps=100,
#         save_steps=100,
#         eval_steps=100,
#         save_total_limit=10,
#         gradient_accumulation_steps=4,
#     )

#     samples_save_dir = "samples_save_dir"
#     batch_size = 8
#     epochs = 10
#     mask_rate = 0.2
#     train_size = 1

#     tabular_rtf = REaLTabFormer(
#         model_type="tabular",
#         samples_save_dir=samples_save_dir,
#         epochs=epochs, batch_size=batch_size,
#         mask_rate=mask_rate,
#         train_size=train_size,
#         random_state=RANDOM_SEED,
#         **training_args_kwargs)



"""Entry point when invoked with python -m realtabformer."""  # pragma: no cover

if __name__ == "__main__":  # pragma: no cover
    import sys

    import realtabformer

    if "--version" in sys.argv:
        print(f"REaLTabFormer version: {realtabformer.__version__}")



# Configuration file for the Sphinx documentation builder.
import os
import sys

from pathlib import Path

# -- Project information
project = "REaLTabFormer"
copyright = "2023, World Bank"
author = "Aivin Solatorio"

# -- General configuration
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "autoapi.extension",
    "myst_nb",
    "sphinx_autodoc_typehints",
]
autoapi_dirs = ["../src"]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
source_suffix = [".rst", ".md"]
templates_path = ["_templates"]

html_logo = "../img/REalTabFormer_Final_EQ.png"
html_title = "REaLTabFormer"

html_theme = "sphinx_book_theme"
html_theme_options = {
    "repository_url": "https://github.com/worldbank/realtabformer",
    "use_repository_button": True,
    "use_issues_button": True,
    "use_edit_page_button": True,
    "logo": {
        "text": f"REaLTabFormer {(Path(__file__).parent.parent / 'src' / 'realtabformer'/ 'VERSION').read_text().strip()}"
    },
}



from pathlib import Path

from . import (
    data_utils,
    realtabformer,
    rtf_analyze,
    rtf_datacollator,
    rtf_exceptions,
    rtf_sampler,
    rtf_trainer,
    rtf_validators,
)
from .realtabformer import REaLTabFormer

__version__ = (Path(__file__).parent / "VERSION").read_text().strip()

__all__ = [
    "REaLTabFormer",
    "realtabformer",
    "data_utils",
    "rtf_sampler",
    "rtf_validators",
    "rtf_exceptions",
    "rtf_datacollator",
    "rtf_trainer",
    "rtf_analyze",
]



"""This module contains some common functions that will be
used across the different tests.
"""
from pathlib import Path
import pandas as pd
import torch
from realtabformer import data_utils as du
from realtabformer import REaLTabFormer

TEST_DATA_DIR = Path(__file__).parent / "data"
TEST_DATA_DIR.mkdir(parents=True, exist_ok=True)


def get_test_data():
    """Get or load the titanic dataset."""
    data_fname = TEST_DATA_DIR / "titanic.csv"
    data_fname.parent.mkdir(parents=True, exist_ok=True)

    if not data_fname.exists():
        data = pd.read_csv("https://raw.githubusercontent.com/datasciencedojo/datasets/master/titanic.csv", index_col=0)
        data.to_csv(data_fname)
    else:
        data = pd.read_csv(data_fname, index_col=0)

    data = data.drop("Name", axis=1)

    return data


def tabular_fixtures(fit=False):
    """Generates tabular fixtures for the sampler test."""
    df = get_test_data().head(100)

    rtf_model = REaLTabFormer(
        model_type=du.ModelType.tabular,
        epochs=1,
        checkpoints_dir=TEST_DATA_DIR / "rtf_checkpoints",
        samples_save_dir=TEST_DATA_DIR / "rtf_smaples",
    )

    if fit:
        rtf_model.fit(df, device="cuda" if torch.cuda.is_available() else "cpu")

    return dict(
        rtf_model=rtf_model,
        df=df,
    )



"""The REaLTabFormer implements the model training and data processing
for tabular and relational data.
"""
import json
import logging
import math
import os
import random
import shutil
import time
import warnings
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
import torch
from datasets import Dataset
from sklearn.metrics.pairwise import manhattan_distances

# from sklearn.metrics import accuracy_score
from transformers import (
    EarlyStoppingCallback,
    EncoderDecoderConfig,
    EncoderDecoderModel,
    PreTrainedModel,
    Seq2SeqTrainer,
    Seq2SeqTrainingArguments,
    Trainer,
    TrainingArguments,
)
from transformers.models.gpt2 import GPT2Config, GPT2LMHeadModel

import realtabformer

from .data_utils import (
    ModelFileName,
    ModelType,
    SpecialTokens,
    TabularArtefact,
    build_vocab,
    make_dataset,
    make_relational_dataset,
    process_data,
)
from .rtf_analyze import SyntheticDataBench
from .rtf_datacollator import RelationalDataCollator
from .rtf_exceptions import SampleEmptyLimitError
from .rtf_sampler import RelationalSampler, TabularSampler
from .rtf_trainer import ResumableTrainer
from .rtf_validators import ObservationValidator


def _normalize_gpt2_state_dict(state_dict):
    state = []
    for key, value in state_dict.items():
        if key.startswith("transformer."):
            # The saved state prefixes the weight names
            # with `transformer.` whereas the
            # encoder expects the weight names to not
            # have the prefix.
            key = key.replace("transformer.", "")

        state.append((key, value))

    return OrderedDict(state)


def _validate_get_device(device: str) -> str:
    if (device == "cuda") and (torch.cuda.device_count() == 0):
        if torch.backends.mps.is_available():
            _device = "mps"
        else:
            _device = "cpu"

        warnings.warn(
            f"The device={device} is not available, using device={_device} instead."
        )
        device = _device

    return device


class REaLTabFormer:
    def __init__(
        self,
        model_type: str,
        tabular_config: Optional[GPT2Config] = None,
        relational_config: Optional[EncoderDecoderConfig] = None,
        parent_realtabformer_path: Optional[Path] = None,
        freeze_parent_model: Optional[bool] = True,
        checkpoints_dir: str = "rtf_checkpoints",
        samples_save_dir: str = "rtf_samples",
        epochs: int = 100,
        batch_size: int = 8,
        random_state: int = 1029,
        train_size: float = 1,
        output_max_length: int = 512,
        early_stopping_patience: int = 5,
        early_stopping_threshold: float = 0,
        mask_rate: float = 0,
        numeric_nparts: int = 1,
        numeric_precision: int = 4,
        numeric_max_len: int = 10,
        **training_args_kwargs,
    ) -> None:
        """Set up a REaLTabFormer instance.

        Args:
            model_type: Explicit declaration of which type of model will be used.
                Can take `tabular` and `relational` as valid values.
            tabular_config: GPT2Config instance to customize the GPT2 model for tabular data
                generation.
            relational_config: EncoderDecoderConfig instance that defines the encoder and decoder
                configs for the encoder-decoder model used for the relational data generation. See
                link for example: https://huggingface.co/docs/transformers/model_doc/encoder-decoder
            parent_realtabformer_path: Path to a saved tabular REaLTabFormer model trained on the
                parent table of a relational tabular data.
            freeze_parent_model: Boolean flag indicating whether the parent-based encoder will be
                frozen or not.
            checkpoints_dir:  Directory where the training checkpoints will be saved
            samples_save_dir: Save the samples generated by this model in this directory.
            epochs: Number of epochs for training the GPT2LM model
            batch_size: Batch size used for training. Must be adjusted based on the available
                compute resource. TrainingArguments is set to use `gradient_accumulation_steps=4`
                which will have an effective batch_size of 32 for the default value.
            train_size: Fraction of the data that will be passed to the `.fit` method that will
                  be used for training. The remaining will be used as validation data.
            output_max_length: Truncation length for the number of output token ids in the
                relational model. This limit applies to the processed data and not the raw number
                of variables. This is not used in the tabular data model.
            early_stopping_patience: Number of evaluation rounds without improvement before
                stopping the training.
            early_stopping_threshold: See link
                https://huggingface.co/docs/transformers/main_classes/callback#transformers.EarlyStoppingCallback.early_stopping_threshold(float,
            mask_rate: The rate of tokens in the transformed observation that will be replaced
                with the [RMASK] token for regularization during training.
            training_args_kwargs: Keyword arguments for the `TrainingArguments` used in training
                the model. Arguments such as `output_dir`, `num_train_epochs`,
                `per_device_train_batch_size`, `per_device_eval_batch_size` if passed will be
                replaced by `checkpoints_dir`, `epochs`, `batch_size arguments`. The comprehensive
                set of options can be found in
              https://huggingface.co/docs/transformers/main/en/main_classes/trainer#transformers.TrainingArguments
        """
        self.model: PreTrainedModel = None

        # This will be set during and will also be deleted after training.
        self.dataset = None

        if model_type not in ModelType.types():
            self._invalid_model_type(model_type)

        self.model_type = model_type

        if self.model_type == ModelType.tabular:
            self._init_tabular(tabular_config)
        elif self.model_type == ModelType.relational:
            self.parent_vocab = None
            self.parent_gpt2_config = None
            self.parent_gpt2_state_dict = None
            self.parent_col_transform_data = None
            self.freeze_parent_model = freeze_parent_model

            if parent_realtabformer_path is not None:
                parent_realtabformer_path = Path(parent_realtabformer_path)

                parent_config = json.loads(
                    (
                        parent_realtabformer_path / ModelFileName.rtf_config_json
                    ).read_text()
                )
                self.parent_col_transform_data = parent_config["col_transform_data"]
                self.parent_vocab = parent_config["vocab"]
                self.parent_gpt2_config = parent_config["tabular_config"]
                self.parent_gpt2_state_dict = _normalize_gpt2_state_dict(
                    torch.load(parent_realtabformer_path / ModelFileName.rtf_model_pt)
                )

            if output_max_length is None:
                warnings.warn(
                    "The `output_max_length` is None. This could result to extended model training if the output length has large variations."
                )

            self.output_max_length = output_max_length
            self._init_relational(relational_config)
        else:
            self._invalid_model_type(self.model_type)

        self.checkpoints_dir = Path(checkpoints_dir)
        self.samples_save_dir = Path(samples_save_dir)
        self.epochs = epochs
        self.batch_size = batch_size

        self.early_stopping_patience = early_stopping_patience
        self.early_stopping_threshold = early_stopping_threshold

        self.training_args_kwargs = dict(
            evaluation_strategy="steps",
            output_dir=self.checkpoints_dir.as_posix(),
            metric_for_best_model="loss",  # This will be replaced with "eval_loss" if `train_size` < 1
            overwrite_output_dir=True,
            num_train_epochs=self.epochs,
            per_device_train_batch_size=self.batch_size,
            per_device_eval_batch_size=self.batch_size,
            gradient_accumulation_steps=4,
            remove_unused_columns=True,
            logging_steps=100,
            save_steps=100,
            eval_steps=100,
            load_best_model_at_end=True,
            save_total_limit=early_stopping_patience + 1,
            optim="adamw_torch",
        )

        # Remove experiment params from `training_args_kwargs`
        for p in [
            "output_dir",
            "num_train_epochs",
            "per_device_train_batch_size",
            "per_device_eval_batch_size",
        ]:
            if p in training_args_kwargs:
                warnings.warn(
                    f"Argument {p} was passed in training_args_kwargs but will be ignored..."
                )
                training_args_kwargs.pop(p)

        self.training_args_kwargs.update(training_args_kwargs)

        self.train_size = train_size
        self.mask_rate = mask_rate

        self.columns: List[str] = []
        self.column_dtypes: Dict[str, type] = {}
        self.column_has_missing: Dict[str, bool] = {}
        self.drop_na_cols: List[str] = []
        self.processed_columns: List[str] = []
        self.numeric_columns: List[str] = []
        self.datetime_columns: List[str] = []
        self.vocab: Dict[str, dict] = {}
        # Output length for generator model
        # including special tokens.
        self.tabular_max_length = None
        self.relational_max_length = None
        # Number of derived columns for the relational
        # and tabular data after performing the data transformation.
        # This will be used as record size validator in the
        # sampling stage.
        self.tabular_col_size = None
        self.relational_col_size = None

        # This stores the transformation
        # parameters for numeric columns.
        self.col_transform_data: Optional[Dict] = None

        # This is the col_transform_data
        # for the relational models's in_df.
        self.in_col_transform_data: Optional[Dict] = None

        self.col_idx_ids: Dict[int, list] = {}

        self.random_state = random_state

        self.numeric_nparts = numeric_nparts
        self.numeric_precision = numeric_precision
        self.numeric_max_len = numeric_max_len

        # A unique identifier for the experiment set after the
        # model is trained.
        self.experiment_id = None
        self.trainer_state = None

        # Target column, when set, a copy of the column values will be
        # implicitly placed at the beginning of the dataframe.
        self.target_col = None

        self.realtabformer_version = realtabformer.__version__

    def _invalid_model_type(self, model_type):
        raise ValueError(
            f"Model type: {model_type} is not valid. REaLTabFormer only supports \
                `tabular` and `relational` values."
        )

    def _init_tabular(self, tabular_config):
        if tabular_config is not None:
            warnings.warn(
                "The `bos_token_id`, `eos_token_id`, and `vocab_size` attributes will \
                    be replaced when the `.fit` method is run."
            )
        else:
            # Default is 12, use 6 for distill-gpt2 as default
            tabular_config = GPT2Config(n_layer=6)

        self.tabular_config = tabular_config
        self.model = None

    def _init_relational(self, relational_config):
        if relational_config is not None:
            warnings.warn(
                "The `bos_token_id`, `eos_token_id`, and `vocab_size` attributes for the \
                    encoder and decoder will be replaced when the `.fit` method is run."
            )
        else:
            # Default is 12, use 6 for distill-gpt2 as default
            relational_config = EncoderDecoderConfig(
                encoder=GPT2Config(n_layer=6).to_dict(),
                decoder=GPT2Config(n_layer=6).to_dict(),
            )

        if self.parent_gpt2_config is not None:
            warnings.warn(
                "A trained model for the parent table is available. The encoder will use the \
                    pretrained config and weights."
            )
            relational_config.encoder = GPT2Config(**self.parent_gpt2_config)

        self.relational_config = relational_config
        self.model = None

    def _extract_column_info(self, df: pd.DataFrame) -> None:
        # Track the column order of the original data
        self.columns = df.columns.to_list()

        # Store the dtypes of the columns
        self.column_dtypes = df.dtypes.astype(str).to_dict()

        # Track which variables have missing values
        self.column_has_missing = (df.isnull().sum() > 0).to_dict()

        # Get the columns where there should be no missing values
        self.drop_na_cols = [
            col for col, has_na in self.column_has_missing.items() if not has_na
        ]

        # Identify the numeric columns. These will undergo
        # special preprocessing.
        self.numeric_columns = df.select_dtypes(include=np.number).columns.to_list()

        # Identify the datetime columns. These will undergo
        # special preprocessing.
        self.datetime_columns = df.select_dtypes(include="datetime").columns.to_list()

    def _generate_vocab(self, df: pd.DataFrame) -> dict:
        return build_vocab(df, special_tokens=SpecialTokens.tokens(), add_columns=False)

    def _check_model(self):
        assert self.model is not None, "Model is None. Train the model first!"

    def _split_train_eval_dataset(self, dataset: Dataset):
        test_size = 1 - self.train_size
        if test_size > 0:
            dataset = dataset.train_test_split(
                test_size=test_size, seed=self.random_state
            )
            dataset["train_dataset"] = dataset.pop("train")
            dataset["eval_dataset"] = dataset.pop("test")

            # Override `metric_for_best_model` from "loss" to "eval_loss"
            self.training_args_kwargs["metric_for_best_model"] = "eval_loss"
            # Make this explicit so that no assumption is made on the
            # direction of the metric improvement.
            self.training_args_kwargs["greater_is_better"] = False
        else:
            dataset = dict(train_dataset=dataset)
            self.training_args_kwargs["evaluation_strategy"] = "no"
            self.training_args_kwargs["load_best_model_at_end"] = False

        return dataset

    def fit(
        self,
        df: pd.DataFrame,
        in_df: Optional[pd.DataFrame] = None,
        join_on: Optional[str] = None,
        resume_from_checkpoint: Union[bool, str] = False,
        device="cuda",
        num_bootstrap: int = 500,
        frac: float = 0.165,
        frac_max_data: int = 10000,
        qt_max: Union[str, float] = 0.05,
        qt_max_default: float = 0.05,
        qt_interval: int = 100,
        qt_interval_unique: int = 100,
        distance: manhattan_distances = manhattan_distances,
        quantile: float = 0.95,
        n_critic: int = 5,
        n_critic_stop: int = 2,
        gen_rounds: int = 3,
        sensitivity_max_col_nums: int = 20,
        use_ks: bool = False,
        full_sensitivity: bool = False,
        sensitivity_orig_frac_multiple: int = 4,
        orig_samples_rounds: int = 5,
        load_from_best_mean_sensitivity: bool = False,
        target_col: str = None,
    ):
        """Train the REaLTabFormer model on the tabular data.

        Args:
            df: Pandas DataFrame containing the tabular data that will be generated during sampling.
              This data goes to the decoder for the relational model.
            in_df: Pandas DataFrame containing observations related to `df`, and from which the
              model will generate data. This data goes to the encoder for the relational model.
            join_on: Column name that links the `df` and the `in_df` tables.
            resume_from_checkpoint: If True, resumes training from the latest checkpoint in the
              checkpoints_dir. If path, resumes the training from the given checkpoint.
            device: Device where the model and the training will be run.
              Use torch devices, e.g., `cpu`, `cuda`, `mps` (experimental)
            num_bootstrap: Number of Bootstrap samples
            frac: The fraction of the data used for training.
            frac_max_data: The maximum number of rows that the training data will have.
            qt_max: The maximum quantile for the discriminator.
            qt_max_default: The default maximum quantile for the discriminator.
            qt_interval: Interval for the quantile check during the training process.
            qt_interval_unique: Interval for the quantile check during the training process.
            distance: Distance metric used for discriminator.
            quantile: The quantile value that the discriminator will be trained to.
            n_critic: Interval between epochs to perform a discriminator assessment.
            n_critic_stop: The number of critic rounds without improvement after which the training
              will be stopped.
            gen_rounds: The number of generator rounds.
            sensitivity_max_col_nums: The maximum number of columns used to compute sensitivity.
            use_ks: Whether to use KS test or not.
            full_sensitivity: Whether to use full sensitivity or not.
            sensitivity_orig_frac_multiple: The size of the training data relative to the chosen
              `frac` that will be used in computing the sensitivity. The larger this value is, the
              more robust the sensitivity threshold will be. However,
              `(sensitivity_orig_frac_multiple + 2)` multiplied by `frac` must be less than 1.
            orig_samples_rounds: This is the number of train/hold-out samples that will be used to
              compute the epoch sensitivity value.
            load_from_best_mean_sensitivity: Whether to load from best mean sensitivity or not.
            target_col: The target column name.

        Returns:
            Trainer
        """
        device = _validate_get_device(device)

        # Set target col for teacher forcing
        self.target_col = target_col

        # Set the seed for, *hopefully*, replicability.
        # This may cause an unexpected behavior when using
        # the resume_from_checkpoint option.
        if self.random_state:
            random.seed(self.random_state)
            np.random.seed(self.random_state)
            torch.manual_seed(self.random_state)
            torch.cuda.manual_seed_all(self.random_state)

        if self.model_type == ModelType.tabular:
            if n_critic <= 0:
                trainer = self._fit_tabular(df, device=device)
                trainer.train(resume_from_checkpoint=resume_from_checkpoint)
            else:
                trainer = self._train_with_sensitivity(
                    df,
                    device,
                    num_bootstrap=num_bootstrap,
                    frac=frac,
                    frac_max_data=frac_max_data,
                    qt_max=qt_max,
                    qt_max_default=qt_max_default,
                    qt_interval=qt_interval,
                    qt_interval_unique=qt_interval_unique,
                    distance=distance,
                    quantile=quantile,
                    n_critic=n_critic,
                    n_critic_stop=n_critic_stop,
                    gen_rounds=gen_rounds,
                    resume_from_checkpoint=resume_from_checkpoint,
                    sensitivity_max_col_nums=sensitivity_max_col_nums,
                    use_ks=use_ks,
                    full_sensitivity=full_sensitivity,
                    sensitivity_orig_frac_multiple=sensitivity_orig_frac_multiple,
                    orig_samples_rounds=orig_samples_rounds,
                    load_from_best_mean_sensitivity=load_from_best_mean_sensitivity,
                )

            del self.dataset

        elif self.model_type == ModelType.relational:
            assert (
                in_df is not None
            ), "The REaLTabFormer for relational data requires two tables for training."
            assert join_on is not None, "The column to join the data must not be None."

            trainer = self._fit_relational(df, in_df, join_on=join_on, device=device)
            trainer.train(resume_from_checkpoint=resume_from_checkpoint)
        else:
            self._invalid_model_type(self.model_type)

        try:
            self.experiment_id = f"id{int((time.time() * 10 ** 10)):024}"
            torch.cuda.empty_cache()

            return trainer
        except Exception as exception:
            if device == torch.device("cuda"):
                del self.model
                torch.cuda.empty_cache()
                self.model = None

            raise exception

    def _train_with_sensitivity(
        self,
        df: pd.DataFrame,
        device: str = "cuda",
        num_bootstrap: int = 500,
        frac: float = 0.165,
        frac_max_data: int = 10000,
        qt_max: Union[str, float] = 0.05,
        qt_max_default: float = 0.05,
        qt_interval: int = 100,
        qt_interval_unique: int = 100,
        distance: manhattan_distances = manhattan_distances,
        quantile: float = 0.95,
        n_critic: int = 5,
        n_critic_stop: int = 2,
        gen_rounds: int = 3,
        sensitivity_max_col_nums: int = 20,
        use_ks: bool = False,
        resume_from_checkpoint: Union[bool, str] = False,
        full_sensitivity: bool = False,
        sensitivity_orig_frac_multiple: int = 4,
        orig_samples_rounds: int = 5,
        load_from_best_mean_sensitivity: bool = False,
    ):
        assert gen_rounds >= 1

        _frac = min(frac, frac_max_data / len(df))
        if frac != _frac:
            warnings.warn(
                f"The frac ({frac}) set results to a sample larger than \
                    frac_max_data={frac_max_data}. Setting frac to {_frac}."
            )
            frac = _frac

        trainer: Trainer = None
        dup_rate = df.duplicated().mean()

        if isinstance(qt_max, str):
            if qt_max == "compute":
                # The idea behind this is if the empirical has
                # natural duplicates, we can use that as
                # basis for what a typical rate for duplicates a
                # random sample should have. Any signidican excess
                # from this indicates overfitting.
                # The choice of dividing the duplicate rate by 2
                # is arbitrary but reasonable to prevent delayed
                # stopping when overfitting.
                dup_rate = dup_rate / 2
                qt_max = dup_rate if dup_rate > 0 else qt_max_default
            else:
                raise ValueError(f"Unexpected qt_max value: {qt_max}")
        elif not isinstance(qt_max, str) and dup_rate >= qt_max:
            warnings.warn(
                f'The qt_max ({qt_max}) set is lower than the duplicate \rate ({dup_rate}) in \
                    the data. This will not give a reliable early stopping condition. Consider \
                        using qt_max="compute" argument.'
            )

        if dup_rate == 0:
            # We do this because for data without unique values, we
            # expect that a generated sample should have equal likelihood
            # in the minimum distance with the hold out.
            warnings.warn(
                f"Duplicate rate ({dup_rate}) in the data is zero. The `qt_interval` will be set \
                    to qt_interval_unique={qt_interval_unique}."
            )
            qt_interval = qt_interval_unique

        # Estimate the sensitivity threshold
        print("Computing the sensitivity threshold...")

        if not full_sensitivity:
            # Dynamically compute the qt_interval to fit the data
            # if the resulting sample has lower resolution.
            # For example, we can't use qt_interval=1000 if the number
            # of samples left at qt_max of the distance matrix is less than
            # 1000.
            # The formula means:
            # - 2       -> accounts for the fact that we concatenate the rows and columns
            #             of the distance matrix.
            # - frac    -> the proportion of the training data that is used to compute the
            #             the distance matrix.
            # - qt_max  -> the maximum quantile of assessment.
            # We divide by 2 to increase the resolution a bit
            _qt_interval = min(qt_interval, (2 * frac * len(df) * qt_max) // 2)
            _qt_interval = max(_qt_interval, 2)
            _qt_interval = int(_qt_interval)

            if _qt_interval < qt_interval:
                warnings.warn(
                    f"qt_interval adjusted from {qt_interval} to {_qt_interval}..."
                )
                qt_interval = _qt_interval

        # Computing this here before splitting may have some data
        # leakage issue, but it should be almost negligible. Doing
        # the computation of the threshold on the full data with the
        # train size aligned will give a more reliable estimate of
        # the sensitivity threshold.
        sensitivity_values = SyntheticDataBench.compute_sensitivity_threshold(
            train_data=df,
            num_bootstrap=num_bootstrap,
            # Divide by two so that the train_data in this computation matches the size
            # of the final df used to train the model. This is essential so that the
            # sensitivity_threshold value is consistent with the val_sensitivity.
            # Concretely, the computation of the distribution of min distances is
            # relative to the number of training observations.
            # The `frac` in  this method corresponds to the size of both the test and the
            # synthetic samples.
            frac=frac / 2,
            qt_max=qt_max,
            qt_interval=qt_interval,
            distance=distance,
            return_values=True,
            quantile=quantile,
            max_col_nums=sensitivity_max_col_nums,
            use_ks=use_ks,
            full_sensitivity=full_sensitivity,
            sensitivity_orig_frac_multiple=sensitivity_orig_frac_multiple,
        )
        sensitivity_threshold = np.quantile(sensitivity_values, quantile)
        mean_sensitivity_value = np.mean(sensitivity_values)
        best_mean_sensitivity_value = np.inf

        assert isinstance(sensitivity_threshold, float)
        print("Sensitivity threshold:", sensitivity_threshold, "qt_max:", qt_max)

        # # Create a hold out sample for the discriminator model
        # hold_df = df.sample(frac=frac, random_state=self.random_state)
        # df = df.loc[df.index.difference(hold_df.index)]

        # Start training
        logging.info("Start training...")

        # Remove existing checkpoints
        for chkp in self.checkpoints_dir.glob("checkpoint-*"):
            shutil.rmtree(chkp, ignore_errors=True)

        sensitivity_scores = []
        bdm_path = self.checkpoints_dir / TabularArtefact.best_disc_model
        mean_closest_bdm_path = (
            self.checkpoints_dir / TabularArtefact.mean_best_disc_model
        )
        not_bdm_path = self.checkpoints_dir / TabularArtefact.not_best_disc_model
        last_epoch_path = self.checkpoints_dir / TabularArtefact.last_epoch_model

        # Remove existing artefacts in the best model dir
        shutil.rmtree(bdm_path, ignore_errors=True)
        bdm_path.mkdir(parents=True, exist_ok=True)

        shutil.rmtree(mean_closest_bdm_path, ignore_errors=True)
        mean_closest_bdm_path.mkdir(parents=True, exist_ok=True)

        shutil.rmtree(not_bdm_path, ignore_errors=True)
        not_bdm_path.mkdir(parents=True, exist_ok=True)

        shutil.rmtree(last_epoch_path, ignore_errors=True)
        last_epoch_path.mkdir(parents=True, exist_ok=True)

        last_epoch = 0
        not_best_val_sensitivity = np.inf

        if resume_from_checkpoint:
            chkp_list = sorted(
                self.checkpoints_dir.glob("checkpoint-*"), key=os.path.getmtime
            )
            if chkp_list:
                # Get the most recent checkpoint based on
                # creation time.
                chkp = chkp_list[-1]
                trainer_state = json.loads((chkp / "trainer_state.json").read_text())
                last_epoch = math.ceil(trainer_state["epoch"])

                trainer = self._fit_tabular(
                    df,
                    device=device,
                    num_train_epochs=last_epoch,
                    target_epochs=self.epochs,
                )

        np.random.seed(self.random_state)
        random.seed(self.random_state)

        for p_epoch in range(last_epoch, self.epochs, n_critic):
            gen_total = int(len(df) * frac)

            num_train_epochs = min(p_epoch + n_critic, self.epochs)
            # Perform the discriminator sampling every `n_critic` epochs
            if trainer is None:
                trainer = self._fit_tabular(
                    df,
                    device=device,
                    num_train_epochs=num_train_epochs,
                    target_epochs=self.epochs,
                )
                trainer.train(resume_from_checkpoint=False)
            else:
                trainer = self._build_tabular_trainer(
                    device=device,
                    num_train_epochs=num_train_epochs,
                    target_epochs=self.epochs,
                )
                trainer.train(resume_from_checkpoint=True)

            try:
                # Generate samples
                gen_df = self.sample(n_samples=gen_rounds * gen_total, device=device)
            except SampleEmptyLimitError:
                # Continue training if the model is still not
                # able to generate stable observations.
                continue

            val_sensitivities = []

            if full_sensitivity:
                for _ in range(gen_rounds):
                    hold_df = df.sample(n=gen_total)

                    for g_idx in range(gen_rounds):
                        val_sensitivities.append(
                            SyntheticDataBench.compute_sensitivity_metric(
                                original=df.loc[df.index.difference(hold_df.index)],
                                synthetic=gen_df.iloc[
                                    g_idx * gen_total : (g_idx + 1) * gen_total
                                ],
                                test=hold_df,
                                qt_max=qt_max,
                                qt_interval=qt_interval,
                                distance=distance,
                                max_col_nums=sensitivity_max_col_nums,
                                use_ks=use_ks,
                            )
                        )
            else:
                for g_idx in range(gen_rounds):
                    for _ in range(orig_samples_rounds):
                        original_df = df.sample(
                            n=sensitivity_orig_frac_multiple * gen_total, replace=False
                        )
                        hold_df = df.loc[df.index.difference(original_df.index)].sample(
                            n=gen_total, replace=False
                        )

                        val_sensitivities.append(
                            SyntheticDataBench.compute_sensitivity_metric(
                                original=original_df,
                                synthetic=gen_df.iloc[
                                    g_idx * gen_total : (g_idx + 1) * gen_total
                                ],
                                test=hold_df,
                                qt_max=qt_max,
                                qt_interval=qt_interval,
                                distance=distance,
                                max_col_nums=sensitivity_max_col_nums,
                                use_ks=use_ks,
                            )
                        )

            val_sensitivity = np.mean(val_sensitivities)

            sensitivity_scores.append(val_sensitivity)

            if val_sensitivity < sensitivity_threshold:
                # Just save the model while the
                # validation sensitivity is still within
                # the accepted range.
                # This way we can load the acceptable
                # model back when the threshold is breached.
                trainer.save_model(bdm_path.as_posix())
                trainer.state.save_to_json((bdm_path / "trainer_state.json").as_posix())

            elif not_best_val_sensitivity > (val_sensitivity - sensitivity_threshold):
                print("Saving not-best model...")
                trainer.save_model(not_bdm_path.as_posix())
                trainer.state.save_to_json(
                    (not_bdm_path / "trainer_state.json").as_posix()
                )
                not_best_val_sensitivity = val_sensitivity - sensitivity_threshold

            _delta_mean_sensitivity_value = abs(
                mean_sensitivity_value - val_sensitivity
            )

            if _delta_mean_sensitivity_value < best_mean_sensitivity_value:
                best_mean_sensitivity_value = _delta_mean_sensitivity_value
                trainer.save_model(mean_closest_bdm_path.as_posix())
                trainer.state.save_to_json(
                    (mean_closest_bdm_path / "trainer_state.json").as_posix()
                )

            print(
                f"Critic round: {p_epoch + n_critic}, \
                    sensitivity_threshold: {sensitivity_threshold}, \
                        val_sensitivity: {val_sensitivity}, \
                            val_sensitivities: {val_sensitivities}"
            )

            if len(sensitivity_scores) > n_critic_stop:
                n_no_improve = 0
                for sensitivity_score in sensitivity_scores[-n_critic_stop:]:
                    # We count no improvement if the score is not
                    # better than the best, and that the score is not
                    # better than the previous score.
                    if sensitivity_score > sensitivity_threshold:
                        n_no_improve += 1

                if n_no_improve == n_critic_stop:
                    print("Stopping training, no improvement in critic...")
                    break

        # Save last epoch artefacts before loading the best model.
        trainer.save_model(last_epoch_path.as_posix())
        trainer.state.save_to_json((last_epoch_path / "trainer_state.json").as_posix())

        loaded_model_path = None

        if not load_from_best_mean_sensitivity:
            if (bdm_path / "pytorch_model.bin").exists():
                loaded_model_path = bdm_path
        else:
            if (mean_closest_bdm_path / "pytorch_model.bin").exists():
                loaded_model_path = mean_closest_bdm_path

        if loaded_model_path is None:
            # There should always be at least one `mean_closest_bdm_path` but
            # in case it doesn't exist, try loading from `not_bdm_path`.
            warnings.warn(
                "No best model was saved. Loading the closest model to the sensitivity_threshold."
            )
            loaded_model_path = not_bdm_path

        self.model = self.model.from_pretrained(loaded_model_path.as_posix())
        self.trainer_state = json.loads(
            (loaded_model_path / "trainer_state.json").read_text()
        )

        return trainer

    def _set_up_relational_coder_configs(self) -> None:
        def _get_coder(coder_name) -> GPT2Config:
            return getattr(self.relational_config, coder_name)

        for coder_name in ["encoder", "decoder"]:
            coder = _get_coder(coder_name)

            coder.bos_token_id = self.vocab[coder_name]["token2id"][SpecialTokens.BOS]
            coder.eos_token_id = self.vocab[coder_name]["token2id"][SpecialTokens.EOS]
            coder.pad_token_id = self.vocab[coder_name]["token2id"][SpecialTokens.PAD]
            coder.vocab_size = len(self.vocab[coder_name]["id2token"])

            if coder_name == "decoder":
                self.relational_config.bos_token_id = coder.bos_token_id
                self.relational_config.eos_token_id = coder.eos_token_id
                self.relational_config.pad_token_id = coder.pad_token_id
                self.relational_config.decoder_start_token_id = coder.eos_token_id

            # Make sure that we have at least the number of
            # columns in the transformed data as positions.
            # This will prevent runtime error.
            # `RuntimeError: CUDA error: device-side assert triggered`
            assert self.relational_max_length
            if (
                coder_name == "decoder"
                and coder.n_positions < self.relational_max_length
            ):
                coder.n_positions = 128 + self.relational_max_length
            elif coder_name == "encoder" and getattr(
                coder, "n_positions", getattr(coder, "max_position_embeddings")
            ) < len(self.vocab[coder_name]["column_token_ids"]):
                positions = 128 + len(self.vocab[coder_name]["column_token_ids"])
                try:
                    coder.n_positions = positions
                except:
                    coder.max_position_embeddings = positions

        # This must be set to True for the EncoderDecoderModel to work at least
        # with GPT2 as the decoder.
        self.relational_config.decoder.add_cross_attention = True

    def _fit_relational(
        self, out_df: pd.DataFrame, in_df: pd.DataFrame, join_on: str, device="cuda"
    ):
        # bert2bert.config.decoder_start_token_id = tokenizer.cls_token_id
        # bert2bert.config.eos_token_id = tokenizer.sep_token_id
        # bert2bert.config.pad_token_id = tokenizer.pad_token_id
        # bert2bert.config.vocab_size = bert2bert.config.encoder.vocab_size

        # All join values in the out_df must be present in the in_df.
        assert len(set(out_df[join_on].unique()).difference(in_df[join_on])) == 0

        # Get the list of index of observations that are related based on
        # the join_on variable.
        common_out_idx = (
            out_df.reset_index(drop=True)
            .groupby(join_on)
            .apply(lambda x: x.index.to_list())
        )

        # Track the mapping of index from input to the list of output indices.
        in_out_idx = pd.Series(
            # Reset the index so that we are sure that the index ids are set properly.
            dict(in_df[join_on].reset_index(drop=True).items())
        ).map(lambda x: common_out_idx.get(x, []))

        # Remove the unique id column from the in_df and the out_df
        in_df = in_df.drop(join_on, axis=1)
        out_df = out_df.drop(join_on, axis=1)

        self._extract_column_info(out_df)
        out_df, self.col_transform_data = process_data(
            out_df,
            numeric_max_len=self.numeric_max_len,
            numeric_precision=self.numeric_precision,
            numeric_nparts=self.numeric_nparts,
        )
        self.processed_columns = out_df.columns.to_list()
        self.vocab["decoder"] = self._generate_vocab(out_df)
        self.relational_col_size = out_df.shape[1]

        # NOTE: the index starts at zero, but should be adjusted
        # to account for the special tokens. For relational data,
        # the index should start at 3 ([[EOS], [BOS], [BMEM]]).
        self.col_idx_ids = {
            ix: self.vocab["decoder"]["column_token_ids"][col]
            for ix, col in enumerate(self.processed_columns)
        }

        # Add these special tokens at specific key values
        # which are used in `REaLSampler._get_relational_col_idx_ids`
        self.col_idx_ids[-1] = [
            self.vocab["decoder"]["token2id"][SpecialTokens.BMEM],
            self.vocab["decoder"]["token2id"][SpecialTokens.EOS],
        ]
        self.col_idx_ids[-2] = [self.vocab["decoder"]["token2id"][SpecialTokens.EMEM]]

        # TODO: handle the col_transform_data from the in_df as well.
        in_df, self.in_col_transform_data = process_data(
            in_df,
            numeric_max_len=self.numeric_max_len,
            numeric_precision=self.numeric_precision,
            numeric_nparts=self.numeric_nparts,
            col_transform_data=self.parent_col_transform_data,
        )
        if self.parent_vocab is None:
            self.vocab["encoder"] = self._generate_vocab(in_df)
        else:
            self.vocab["encoder"] = self.parent_vocab

        # Load the dataframe into a HuggingFace Dataset
        dataset = make_relational_dataset(
            in_df=in_df,
            out_df=out_df,
            vocab=self.vocab,
            in_out_idx=in_out_idx,
            output_max_length=self.output_max_length,
            mask_rate=self.mask_rate,
            return_token_type_ids=False,
        )

        # Compute the longest sequence of labels in the dataset and add a buffer of 1.
        self.relational_max_length = (
            max(
                dataset.map(lambda example: dict(length=len(example["labels"])))[
                    "length"
                ]
            )
            + 1
        )

        # Create train-eval split if specified
        dataset = self._split_train_eval_dataset(dataset)

        # Set up the config and the model
        self._set_up_relational_coder_configs()

        # Build the model.
        self.model = EncoderDecoderModel(self.relational_config)
        if self.parent_gpt2_state_dict is not None:
            pretrain_load = self.model.encoder.load_state_dict(
                self.parent_gpt2_state_dict, strict=False
            )
            assert (
                not pretrain_load.missing_keys
            ), "There should be no missing_keys after loading the pretrained GPT2 state!"

            if self.freeze_parent_model:
                # We freeze the weights if we use the pretrained
                # parent table model.
                for param in self.model.encoder.parameters():
                    param.requires_grad = False

        # Tell pytorch to run this model on the GPU.
        device = torch.device(device)
        if device == torch.device("cuda"):
            self.model.cuda()

        # Set TrainingArguments and the Seq2SeqTrainer
        training_args_kwargs = dict(self.training_args_kwargs)

        default_args_kwargs = dict(
            # predict_with_generate=True,
            # warmup_steps=2000,
            fp16=(
                device == torch.device("cuda")
            ),  # Use fp16 by default if using cuda device
        )

        for k, v in default_args_kwargs.items():
            if k not in training_args_kwargs:
                training_args_kwargs[k] = v

        callbacks = None
        if training_args_kwargs["load_best_model_at_end"]:
            callbacks = [
                EarlyStoppingCallback(
                    self.early_stopping_patience, self.early_stopping_threshold
                )
            ]

        # instantiate trainer
        trainer = Seq2SeqTrainer(
            model=self.model,
            args=Seq2SeqTrainingArguments(**training_args_kwargs),
            callbacks=callbacks,
            data_collator=RelationalDataCollator(),
            **dataset,
        )

        return trainer

    def _fit_tabular(
        self,
        df: pd.DataFrame,
        device="cuda",
        num_train_epochs: int = None,
        target_epochs: int = None,
    ) -> Trainer:
        self._extract_column_info(df)
        df, self.col_transform_data = process_data(
            df,
            numeric_max_len=self.numeric_max_len,
            numeric_precision=self.numeric_precision,
            numeric_nparts=self.numeric_nparts,
            target_col=self.target_col,
        )
        self.processed_columns = df.columns.to_list()
        self.vocab = self._generate_vocab(df)
        self.tabular_col_size = df.shape[0]

        # NOTE: the index starts at zero, but should be adjusted
        # to account for the special tokens. For tabular data,
        # the index should start at 1.
        self.col_idx_ids = {
            ix: self.vocab["column_token_ids"][col]
            for ix, col in enumerate(self.processed_columns)
        }

        # Load the dataframe into a HuggingFace Dataset
        dataset = make_dataset(
            df, self.vocab, mask_rate=self.mask_rate, return_token_type_ids=False
        )

        # Store the sequence length for the processed data
        self.tabular_max_length = len(dataset[0]["input_ids"])

        # Create train-eval split if specified
        dataset = self._split_train_eval_dataset(dataset)
        self.dataset = dataset

        # Set up the config and the model
        self.tabular_config.bos_token_id = self.vocab["token2id"][SpecialTokens.BOS]
        self.tabular_config.eos_token_id = self.vocab["token2id"][SpecialTokens.EOS]
        self.tabular_config.vocab_size = len(self.vocab["id2token"])

        # Make sure that we have at least the number of
        # columns in the transformed data as positions.
        if self.tabular_config.n_positions < len(self.vocab["column_token_ids"]):
            self.tabular_config.n_positions = 128 + len(self.vocab["column_token_ids"])

        self.model = GPT2LMHeadModel(self.tabular_config)

        # Tell pytorch to run this model on the GPU.
        device = torch.device(device)
        if device == torch.device("cuda"):
            self.model.cuda()

        return self._build_tabular_trainer(
            device=device,
            num_train_epochs=num_train_epochs,
            target_epochs=target_epochs,
        )

    def _build_tabular_trainer(
        self,
        device="cuda",
        num_train_epochs: int = None,
        target_epochs: int = None,
    ) -> Trainer:
        device = torch.device(device)

        # Set TrainingArguments and the Trainer
        logging.info("Set up the TrainingArguments and the Trainer...")
        training_args_kwargs: Dict[str, Any] = dict(self.training_args_kwargs)

        default_args_kwargs = dict(
            fp16=(
                device == torch.device("cuda")
            ),  # Use fp16 by default if using cuda device
        )

        for k, v in default_args_kwargs.items():
            if k not in training_args_kwargs:
                training_args_kwargs[k] = v

        if num_train_epochs is not None:
            training_args_kwargs["num_train_epochs"] = num_train_epochs

        # # NOTE: The `ResumableTrainer` will default to its original
        # # behavior (Trainer) if `target_epochs`` is None.
        # # Set the `target_epochs` to `num_train_epochs` if not specified.
        # if target_epochs is None:
        #     target_epochs = training_args_kwargs.get("num_train_epochs")

        callbacks = None
        if training_args_kwargs["load_best_model_at_end"]:
            callbacks = [
                EarlyStoppingCallback(
                    self.early_stopping_patience, self.early_stopping_threshold
                )
            ]

        assert self.dataset
        trainer = ResumableTrainer(
            target_epochs=target_epochs,
            save_epochs=None,
            model=self.model,
            args=TrainingArguments(**training_args_kwargs),
            data_collator=None,  # Use the default_data_collator
            callbacks=callbacks,
            **self.dataset,
        )

        return trainer

    def sample(
        self,
        n_samples: int = None,
        input_unique_ids: Optional[Union[pd.Series, List]] = None,
        input_df: Optional[pd.DataFrame] = None,
        input_ids: Optional[torch.tensor] = None,
        gen_batch: Optional[int] = 128,
        device: str = "cuda",
        seed_input: Optional[Union[pd.DataFrame, Dict[str, Any]]] = None,
        save_samples: Optional[bool] = False,
        constrain_tokens_gen: Optional[bool] = True,
        validator: Optional[ObservationValidator] = None,
        continuous_empty_limit: int = 10,
        suppress_tokens: Optional[List[int]] = None,
        forced_decoder_ids: Optional[List[List[int]]] = None,
        related_num: Optional[Union[int, List[int]]] = None,
        **generate_kwargs,
    ) -> pd.DataFrame:
        """Generate synthetic tabular data samples

        Args:
            n_samples: Number of synthetic samples to generate for the tabular data.
            input_unique_ids: The unique identifier that will be used to link the input
              data to the generated values when sampling for relational data.
            input_df: Pandas DataFrame containing the tabular input data.
            input_ids: (NOTE: the `input_df` argument is the preferred input)
              The input_ids that conditions the generation of the relational data.
            gen_batch: Controls the batch size of the data generation process. This parameter
              should be adjusted based on the compute resources.
            device: The device used by the generator.
              Use torch devices, e.g., `cpu`, `cuda`, `mps` (experimental)
            seed_input: A dictionary of `col_name:values` for the seed data. Only `col_names`
              that are actually in the first sequence of the training input will be used.
            constrain_tokens_gen: Set whether we impose a constraint at each step of the generation
              limited only to valid tokens for the column.
            validator: An instance of `ObservationValidator` for validating the generated samples.
              The validators are applied to observations only, and don't support inter-observation
              validation. See `ObservationValidator` docs on how to set up a validator.
            continuous_invalid_limit: The sampling will raise an exception if
              `continuous_empty_limit` empty sample batches have been produced continuously. This
              will prevent an infinite loop if the quality of the data generated is not good and
              always produces invalid observations.
            suppress_tokens: (from docs) A list of tokens that will be supressed at generation.
              The SupressTokens logit processor will set their log probs to -inf so that they are
              not sampled. This is a useful feature for imputing missing values.
            forced_decoder_ids: (from docs) A list of pairs of integers which indicates a mapping
              from generation indices to token indices that will be forced before sampling. For
              example, [[1, 123]] means the second generated token will always be a token of
              index 123. This is a useful feature for constraining the model to generate only
              specific stratification variables in surveys, e.g., GEO1, URBAN/RURAL variables.
            related_num: A column name in the input_df containing the number of observations that the child
             table is expected to have for the parent observation. It can also be an integer if the input_df
             corresponds to a set of observations having the same number of expected observations.
             This parameter is only valid for the relational model.
            generate_kwargs: Additional keywords arguments that will be supplied to `.generate`
              method. For a comprehensive list of arguments, see:
              https://huggingface.co/docs/transformers/v4.24.0/en/main_classes/text_generation#transformers.generation_utils.GenerationMixin.generate

        Returns:
            DataFrame with n_samples rows of generated data
        """
        self._check_model()
        device = _validate_get_device(device)

        # Clear the cache
        torch.cuda.empty_cache()

        if self.model_type == ModelType.tabular:
            assert n_samples
            assert self.tabular_max_length is not None
            assert self.tabular_col_size is not None
            assert self.col_transform_data is not None

            tabular_sampler = TabularSampler.sampler_from_model(
                rtf_model=self, device=device
            )
            # (
            #     model_type=self.model_type,
            #     model=self.model,
            #     vocab=self.vocab,
            #     processed_columns=self.processed_columns,
            #     max_length=self.tabular_max_length,
            #     col_size=self.tabular_col_size,
            #     col_idx_ids=self.col_idx_ids,
            #     columns=self.columns,
            #     datetime_columns=self.datetime_columns,
            #     column_dtypes=self.column_dtypes,
            #     drop_na_cols=self.drop_na_cols,
            #     col_transform_data=self.col_transform_data,
            #     random_state=self.random_state,
            #     device=device,
            # )
            synth_df = tabular_sampler.sample_tabular(
                n_samples=n_samples,
                gen_batch=gen_batch,
                device=device,
                seed_input=seed_input,
                constrain_tokens_gen=constrain_tokens_gen,
                validator=validator,
                continuous_empty_limit=continuous_empty_limit,
                suppress_tokens=suppress_tokens,
                forced_decoder_ids=forced_decoder_ids,
                **generate_kwargs,
            )

        elif self.model_type == ModelType.relational:
            assert (input_ids is not None) or (input_df is not None)
            assert self.relational_max_length is not None
            assert self.relational_col_size is not None
            assert self.col_transform_data is not None
            assert self.in_col_transform_data is not None

            relational_sampler = RelationalSampler.sampler_from_model(
                rtf_model=self, device=device
            )
            # (
            #     model_type=self.model_type,
            #     model=self.model,
            #     vocab=self.vocab,
            #     processed_columns=self.processed_columns,
            #     max_length=self.relational_max_length,
            #     col_size=self.relational_col_size,
            #     col_idx_ids=self.col_idx_ids,
            #     columns=self.columns,
            #     datetime_columns=self.datetime_columns,
            #     column_dtypes=self.column_dtypes,
            #     drop_na_cols=self.drop_na_cols,
            #     col_transform_data=self.col_transform_data,
            #     in_col_transform_data=self.in_col_transform_data,
            #     random_state=self.random_state,
            #     device=device,
            # )

            synth_df = relational_sampler.sample_relational(
                input_unique_ids=input_unique_ids,
                input_df=input_df,
                input_ids=input_ids,
                device=device,
                gen_batch=gen_batch,
                constrain_tokens_gen=constrain_tokens_gen,
                validator=validator,
                continuous_empty_limit=continuous_empty_limit,
                suppress_tokens=suppress_tokens,
                forced_decoder_ids=forced_decoder_ids,
                related_num=related_num,
                **generate_kwargs,
            )

        if save_samples:
            samples_fname = (
                self.samples_save_dir
                / f"rtf_{self.model_type}-exp_{self.experiment_id}-{int(time.time())}-samples_{synth_df.shape[0]}.pkl"
            )
            samples_fname.parent.mkdir(parents=True, exist_ok=True)
            synth_df.to_pickle(samples_fname)

        return synth_df

    def predict(
        self,
        data: pd.DataFrame,
        target_col: str,
        target_pos_val: Any = None,
        batch: int = 32,
        obs_sample: int = 30,
        fillunk: bool = True,
        device: str = "cuda",
        disable_progress_bar: bool = True,
        **generate_kwargs,
    ) -> pd.Series:
        """
        Use the trained model to make predictions on a given dataframe.

        Args:
            data: The data to make predictions on, in the form of a Pandas dataframe.
            target_col: The name of the target column in the data to predict.
            target_pos_val: The positive value in the target column to use for binary
              classification. This is produces a one-to-many prediction relative to
              `target_pos_val` for targets that are multi-categorical.
            batch: The batch size to use when making predictions.
            obs_sample: The number of observations to sample from the data when making predictions.
            fillunk: If True, the function will fill any missing values in the data before making
              predictions. Fill unknown tokens with the mode of the batch in the given step.
            device: The device to use for prediction. Can be either "cpu" or "cuda".
            **generate_kwargs: Additional keyword arguments to pass to the model's `generate`
              method.

        Returns:
            A Pandas series containing the predicted values for the target column.
        """

        assert (
            self.model_type == ModelType.tabular
        ), "The predict method is only implemented for tabular data..."
        self._check_model()
        device = _validate_get_device(device)
        batch = min(batch, data.shape[0])

        # Clear the cache
        torch.cuda.empty_cache()

        # assert self.tabular_max_length is not None
        # assert self.tabular_col_size is not None
        # assert self.col_transform_data is not None

        tabular_sampler = TabularSampler.sampler_from_model(self, device=device)

        # TabularSampler(
        #     model_type=self.model_type,
        #     model=self.model,
        #     vocab=self.vocab,
        #     processed_columns=self.processed_columns,
        #     max_length=self.tabular_max_length,
        #     col_size=self.tabular_col_size,
        #     col_idx_ids=self.col_idx_ids,
        #     columns=self.columns,
        #     datetime_columns=self.datetime_columns,
        #     column_dtypes=self.column_dtypes,
        #     drop_na_cols=self.drop_na_cols,
        #     col_transform_data=self.col_transform_data,
        #     random_state=self.random_state,
        #     device=device,
        # )

        return tabular_sampler.predict(
            data=data,
            target_col=target_col,
            target_pos_val=target_pos_val,
            batch=batch,
            obs_sample=obs_sample,
            fillunk=fillunk,
            device=device,
            disable_progress_bar=disable_progress_bar,
            **generate_kwargs,
        )

    def save(self, path: Union[str, Path], allow_overwrite: Optional[bool] = False):
        """Save REaLTabFormer Model

        Saves the model weights and a configuration file in the given directory.
        Args:
            path: Path where to save the model
        """
        self._check_model()
        assert self.experiment_id is not None

        if isinstance(path, str):
            path = Path(path)

        # Add experiment id to the save path
        path = path / self.experiment_id

        config_file = path / ModelFileName.rtf_config_json
        model_file = path / ModelFileName.rtf_model_pt

        if path.is_dir() and not allow_overwrite:
            if config_file.exists() or model_file.exists():
                raise ValueError(
                    "This directory is not empty, and contains either a config or a model."
                    " Consider setting `allow_overwrite=True` if you want to overwrite these."
                )
            else:
                warnings.warn(
                    f"Directory {path} exists, but `allow_overwrite=False`."
                    " This will raise an error next time when the model artifacts \
                        exist on this directory"
                )

        path.mkdir(parents=True, exist_ok=True)

        # Save attributes
        rtf_attrs = self.__dict__.copy()
        rtf_attrs.pop("model")

        # We don't need to store the `parent_config`
        # since a saved model should have the weights loaded from
        # the trained model already.
        for ignore_key in [
            "parent_vocab",
            "parent_gpt2_config",
            "parent_gpt2_state_dict",
            "parent_col_transform_data",
        ]:
            if ignore_key in rtf_attrs:
                rtf_attrs.pop(ignore_key)

        # GPT2Config is not JSON serializable, let us manually
        # extract the attributes.
        if rtf_attrs.get("tabular_config"):
            rtf_attrs["tabular_config"] = rtf_attrs["tabular_config"].to_dict()

        if rtf_attrs.get("relational_config"):
            rtf_attrs["relational_config"] = rtf_attrs["relational_config"].to_dict()

        rtf_attrs["checkpoints_dir"] = rtf_attrs["checkpoints_dir"].as_posix()
        rtf_attrs["samples_save_dir"] = rtf_attrs["samples_save_dir"].as_posix()

        config_file.write_text(json.dumps(rtf_attrs))

        # Save model weights
        torch.save(self.model.state_dict(), model_file.as_posix())

        if self.model_type == ModelType.tabular:
            # Copy the special model checkpoints for
            # tabular models.
            for artefact in TabularArtefact.artefacts():
                print("Copying artefacts from:", artefact)
                if (self.checkpoints_dir / artefact).exists():
                    shutil.copytree(
                        self.checkpoints_dir / artefact,
                        path / artefact,
                        dirs_exist_ok=True,
                    )

    @classmethod
    def load_from_dir(cls, path: Union[str, Path]):
        """Load a saved REaLTabFormer model

        Load trained REaLTabFormer model from directory.
        Args:
            path: Directory where REaLTabFormer model is saved
        Returns:
            REaLTabFormer instance
        """

        if isinstance(path, str):
            path = Path(path)

        config_file = path / ModelFileName.rtf_config_json
        model_file = path / ModelFileName.rtf_model_pt

        assert path.is_dir(), f"Directory {path} does not exist."
        assert config_file.exists(), f"Config file {config_file} does not exist."
        assert model_file.exists(), f"Model file {model_file} does not exist."

        # Load the saved attributes
        rtf_attrs = json.loads(config_file.read_text())

        # Create new REaLTabFormer model instance
        try:
            realtf = cls(model_type=rtf_attrs["model_type"])
        except KeyError:
            # Back-compatibility for saved models
            # before the support for relational data
            # was implemented.
            realtf = cls(model_type="tabular")

        # Set all attributes and handle the
        # special case for the GPT2Config.
        for k, v in rtf_attrs.items():
            if k == "gpt_config":
                # Back-compatibility for saved models
                # before the support for relational data
                # was implemented.
                v = GPT2Config.from_dict(v)
                k = "tabular_config"

            elif k == "tabular_config":
                v = GPT2Config.from_dict(v)

            elif k == "relational_config":
                v = EncoderDecoderConfig.from_dict(v)

            elif k in ["checkpoints_dir", "samples_save_dir"]:
                v = Path(v)

            elif k == "vocab":
                if realtf.model_type == ModelType.tabular:
                    # Cast id back to int since JSON converts them to string.
                    v["id2token"] = {int(ii): vv for ii, vv in v["id2token"].items()}
                elif realtf.model_type == ModelType.relational:
                    v["encoder"]["id2token"] = {
                        int(ii): vv for ii, vv in v["encoder"]["id2token"].items()
                    }
                    v["decoder"]["id2token"] = {
                        int(ii): vv for ii, vv in v["decoder"]["id2token"].items()
                    }
                else:
                    raise ValueError(f"Invalid model_type: {realtf.model_type}")

            elif k == "col_idx_ids":
                v = {int(ii): vv for ii, vv in v.items()}

            setattr(realtf, k, v)

        # Implement back-compatibility for REaLTabFormer version < 0.0.1.8.2
        # since the attribute `col_idx_ids` is not implemented before.
        if "col_idx_ids" not in rtf_attrs:
            if realtf.model_type == ModelType.tabular:
                realtf.col_idx_ids = {
                    ix: realtf.vocab["column_token_ids"][col]
                    for ix, col in enumerate(realtf.processed_columns)
                }
            elif realtf.model_type == ModelType.relational:
                # NOTE: the index starts at zero, but should be adjusted
                # to account for the special tokens. For relational data,
                # the index should start at 3 ([[EOS], [BOS], [BMEM]]).
                realtf.col_idx_ids = {
                    ix: realtf.vocab["decoder"]["column_token_ids"][col]
                    for ix, col in enumerate(realtf.processed_columns)
                }

                # Add these special tokens at specific key values
                # which are used in `REaLSampler._get_relational_col_idx_ids`
                realtf.col_idx_ids[-1] = [
                    realtf.vocab["decoder"]["token2id"][SpecialTokens.BMEM],
                    realtf.vocab["decoder"]["token2id"][SpecialTokens.EOS],
                ]
                realtf.col_idx_ids[-2] = [
                    realtf.vocab["decoder"]["token2id"][SpecialTokens.EMEM]
                ]

        # Load model weights
        if realtf.model_type == ModelType.tabular:
            realtf.model = GPT2LMHeadModel(realtf.tabular_config)
        elif realtf.model_type == ModelType.relational:
            realtf.model = EncoderDecoderModel(realtf.relational_config)
        else:
            raise ValueError(f"Invalid model_type: {realtf.model_type}")

        realtf.model.load_state_dict(
            torch.load(model_file.as_posix(), map_location="cpu")
        )

        return realtf



import random
import pandas as pd
import pytest

from realtabformer import data_utils as du

df = pd.DataFrame({
    "float": [1.0, 0.123, 32.2334, 100.32],
    "int": [1, 23, 4, 1232],
    "datetime": [
        pd.Timestamp("20180310"),
        pd.Timestamp("20190310"),
        pd.Timestamp("20200316"),
        pd.Timestamp("20181110")],
    "string": [
        "deep", "learning", "data", "science"]})


def test_SPECIAL_COL_SEP():
    # Make sure that any changes in SPECIAL_COL_SEP
    # is deliberate. So have this test here.
    assert du.SPECIAL_COL_SEP == "___"


def test_ColDataType():
    assert du.ColDataType.types() == ["NUMERIC", "DATETIME", "CATEGORICAL"]


def test_SpecialTokens():
    # The order of the tokens as defined
    # in SpecialTokens matters, so
    # we need to make sure that they
    # don't change accidentaly!
    assert du.SpecialTokens.tokens() == [
        "[UNK]",
        "[SEP]",
        "[PAD]",
        "[CLS]",
        "[MASK]",
        "[BOS]",
        "[EOS]",
        "[BMEM]",
        "[EMEM]",
        "[RMASK]",
        "[SPTYPE]",
    ]


def test_get_input_ids():
    # df = du.process_data(df)
    ddf = df.copy()

    with pytest.raises(AssertionError):
        vocab = du.build_vocab(ddf.astype(str), special_tokens=du.SpecialTokens.tokens(), add_columns=True)

    ddf.columns = [f"{idx}___{dtype}___{col}" for idx, (col, dtype) in enumerate(zip(ddf.columns, ["NUMERIC", "NUMERIC", "DATETIME", "CATEGORICAL"]))]
    vocab = du.build_vocab(ddf.astype(str), special_tokens=du.SpecialTokens.tokens(), add_columns=True)

    example = {
        "0___NUMERIC___float": 32.32, "1___NUMERIC___int": 13,
        "2___DATETIME___datetime": pd.Timestamp("20180310"),
        "3___CATEGORICAL___string": "learning"}

    with pytest.raises(AssertionError):
        # Raise assertion for return_token_type_ids=True
        out = du.get_input_ids(
            example,
            vocab=vocab,
            columns=ddf.columns,
            mask_rate=0,
            return_label_ids=True,
            return_token_type_ids=True,
            affix_bos=True,
            affix_eos=True
        )

    out = du.get_input_ids(
        example,
        vocab=vocab,
        columns=ddf.columns,
        mask_rate=0,
        return_label_ids=True,
        return_token_type_ids=False,
        affix_bos=True,
        affix_eos=True
    )

    assert "label_ids" in out
    assert out["label_ids"] == out["input_ids"]
    assert out["input_ids"][0] == vocab["token2id"][du.SpecialTokens.BOS]
    assert out["input_ids"][-1] == vocab["token2id"][du.SpecialTokens.EOS]

    out = du.get_input_ids(
        example,
        vocab=vocab,
        columns=ddf.columns,
        mask_rate=0,
        return_label_ids=True,
        return_token_type_ids=False,
        affix_bos=True,
        affix_eos=False
    )

    assert "label_ids" in out
    assert out["label_ids"] == out["input_ids"]
    assert out["input_ids"][0] == vocab["token2id"][du.SpecialTokens.BOS]
    assert out["input_ids"][-1] != vocab["token2id"][du.SpecialTokens.EOS]


def test_build_vocab_no_add_columns():
    # Test vocab without special tokens.
    # Convert all data to string for convenience.
    vocab = du.build_vocab(df.astype(str), add_columns=False)
    # Check vocab size
    assert len(vocab["id2token"]) == 16
    assert len(vocab["token2id"]) == 16

    # Check id range
    assert min(vocab["id2token"]) == 0
    assert max(vocab["id2token"]) == 15

    # Check that the token2id and id2token
    # actually are inverse maps of each other.
    for token, t_id in vocab["token2id"].items():
        assert vocab["token2id"][token] == t_id

    # Check values
    assert vocab["id2token"][0] == "0.123"
    assert vocab["id2token"][1] == "1.0"

    # Note that "100.32" goes before "32.2334"
    # because we changed the dtype to string.
    # The sorting applies to the string-converted
    # data.
    assert vocab["id2token"][2] == "100.32"
    assert vocab["id2token"][3] == "32.2334"

    # Sample other tokens.
    assert vocab["id2token"][5] == "1232"
    assert vocab["id2token"][10] == "2019-03-10"
    assert vocab["id2token"][14] == "learning"

    # Check the column token ids output.
    assert vocab["column_token_ids"]["int"] == [4, 5, 6, 7]
    assert vocab["column_token_ids"]["string"] == [12, 13, 14, 15]


def test_build_vocab_add_columns():
    # Test vocab without special tokens.
    # Convert all data to string for convenience.
    ddf = df.copy()
    ddf.columns = [f"{idx}___{dtype}___{col}" for idx, (col, dtype) in enumerate(zip(ddf.columns, ["NUMERIC", "NUMERIC", "DATETIME", "CATEGORICAL"]))]

    vocab = du.build_vocab(ddf.astype(str), add_columns=True)
    # Check vocab size
    assert len(vocab["id2token"]) == (16 + 4)
    assert len(vocab["token2id"]) == (16 + 4)

    # Check id range
    assert min(vocab["id2token"]) == 0
    assert max(vocab["id2token"]) == 19

    # Check column values
    assert vocab["id2token"][16] == "0___NUMERIC"
    assert vocab["id2token"][17] == "1___NUMERIC"
    assert vocab["id2token"][18] == "2___DATETIME"
    assert vocab["id2token"][19] == "3___CATEGORICAL"


def test_process_numeric_data():
    series, transform_data = du.process_numeric_data(
        df["int"], max_len=10, numeric_precision=4)
    expected_out = ["0001", "0023", "0004", "1232"]

    for v, e in zip(series, expected_out):
        assert v == e

    # Make sure that max_len doesn't truncate integral
    # data.
    series, transform_data = du.process_numeric_data(
        df["int"], max_len=2, numeric_precision=4)

    expected_out = ["0001", "0023", "0004", "1232"]

    for v, e in zip(series, expected_out):
        assert v == e

    series, transform_data = du.process_numeric_data(
        df["float"], max_len=5, numeric_precision=4)

    # Note that the value of 0.123 will be truncated
    # to 000.1 because we prioritize the padding at
    # the leading digits of the values before truncating
    # the max length.
    expected_out = ["001.0", "000.1", "032.2", "100.3"]

    for v, e in zip(series, expected_out):
        assert v == e

    # Check that the processing of the
    # data is sensitive to the resolution loss.
    with pytest.raises(AssertionError):
        # This should raise an AssertionError because the
        # desired max_len of 4 will generate a loss in
        # the numeric resolution of the data
        series, transform_data = du.process_numeric_data(
            df["float"], max_len=4, numeric_precision=4)


def test_process_data():
    pr_df, _ = du.process_data(
        df, numeric_max_len=10, numeric_precision=4, numeric_nparts=2)

    # Validate the processed columns
    assert pr_df.shape[1] == 12
    assert pr_df.columns.str.startswith(f"0___{du.ColDataType.NUMERIC}___float").sum() == 4
    assert pr_df.columns.str.startswith(f"1___{du.ColDataType.NUMERIC}___int").sum() == 2
    assert pr_df.columns.str.startswith(f"2___{du.ColDataType.DATETIME}___datetime").sum() == 5
    assert pr_df.columns.str.startswith(f"3___{du.ColDataType.CATEGORICAL}___string").sum() == 1

    # Validate that the columns are properly ordered (default)
    start_idx = 0
    for col in pr_df.columns:
        col_idx = int(col.split(du.SPECIAL_COL_SEP)[0])
        if col_idx != start_idx:
            assert (start_idx + 1) == col_idx
            start_idx = col_idx

    for col in pr_df.columns:
        assert pr_df[col].str.startswith(col).all()


@pytest.mark.parametrize("first_col_type", [None, du.ColDataType.CATEGORICAL, du.ColDataType.NUMERIC])
def test_process_data_first_col_type(first_col_type):
    # Test for categorical first cols
    pr_df, _ = du.process_data(
        df, numeric_max_len=10, numeric_precision=4, numeric_nparts=2,
        first_col_type=first_col_type
    )

    start_idx = 0
    seen_last = False
    for idx, col in enumerate(pr_df.columns):
        if first_col_type is not None:
            if idx == 0:
                # Make sure that our set first_col_type
                # is actually the first column type in the
                # returned data.
                assert first_col_type in col
            elif first_col_type not in col:
                seen_last = True

            if not seen_last:
                if first_col_type == du.ColDataType.CATEGORICAL:
                    assert first_col_type in col
                else:
                    # NUMERIC and DATETIME fall under the same
                    # general numeric category.
                    assert (first_col_type in col) or (du.ColDataType.DATETIME in col)
            else:
                assert first_col_type not in col
        else:
            # If no preferred first_col_type is set,
            # we use the actual order of the input
            # dataframe.
            col_idx = int(col.split(du.SPECIAL_COL_SEP)[0])
            if col_idx != start_idx:
                assert (start_idx + 1) == col_idx
                start_idx = col_idx


def test_decode_processed_column():
    assert du.decode_processed_column(f"0___{du.ColDataType.NUMERIC}___float_00") == "float_00"
    assert du.decode_processed_column(f"0___{du.ColDataType.NUMERIC}___int_01") == "int_01"
    assert du.decode_processed_column(f"0___{du.ColDataType.DATETIME}___datetime_03") == "datetime_03"
    assert du.decode_processed_column(f"0___{du.ColDataType.CATEGORICAL}___string_02") == "string_02"

    # Check if leading numeric prefix is long.
    assert du.decode_processed_column(f"121___{du.ColDataType.CATEGORICAL}___string_02") == "string_02"

    # Check if columns is not in the expected format.
    assert du.decode_processed_column("random_col") == "random_col"


@pytest.mark.parametrize("dtype", du.ColDataType.types() + ["SOMEDTYPE"])
def test_encode_processed_column(dtype):
    idx = 0
    col = "foo"

    if dtype != "SOMEDTYPE":
        # Expected form: "0___(NUMERIC|DATETIME|CATEGORICAL)___foo"
        out = f"{idx}{du.SPECIAL_COL_SEP}{dtype}{du.SPECIAL_COL_SEP}{col}"
        assert du.encode_processed_column(idx, dtype, col) == out
    else:
        with pytest.raises(AssertionError):
            du.encode_processed_column(idx, dtype, col)


@pytest.mark.parametrize("dtype", du.ColDataType.types() + ["SOMEDTYPE"])
def test_extract_processed_column(dtype):
    col = "foo012_%$@&#"

    for _ in range(10):
        idx = random.randint(0, 1000)

        if dtype != "SOMEDTYPE":
            # Expected form: "0___(NUMERIC|DATETIME|CATEGORICAL)___foo012_%$@&#"
            expected = f"{idx}{du.SPECIAL_COL_SEP}{dtype}"
            inp = f"{idx}{du.SPECIAL_COL_SEP}{dtype}{du.SPECIAL_COL_SEP}{col}{du.SPECIAL_COL_SEP}00"
            assert du.extract_processed_column(inp) == expected
        else:
            inp = f"{idx}{du.SPECIAL_COL_SEP}{dtype}{du.SPECIAL_COL_SEP}{col}{du.SPECIAL_COL_SEP}00"
            assert du.extract_processed_column(inp) is None



class SampleEmptyError(Exception):
    """Exception raised for generated samples without valid observations.

    Attributes:
        salary -- input salary which caused the error
        message -- explanation of the error
    """

    def __init__(
        self, message="Generated sample is empty after validation.", in_size=None
    ):
        self.in_size = in_size
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.in_size} -> {self.message}"


class SampleEmptyLimitError(SampleEmptyError):
    """Exception raised when SampleEmptyError is raised
    continuously for some specific limit."""

    def __init__(
        self,
        message="Generated sample is still empty after the set limit.",
        in_size=None,
    ):
        super().__init__(message, in_size)



import logging
import random
import re
import time
import uuid
import warnings
from dataclasses import dataclass, fields
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from datasets import Dataset

TEACHER_FORCING_PRE = "_TEACHERFORCING"
SPECIAL_COL_SEP = "___"
NUMERIC_NA_TOKEN = "@"
INVALID_NUMS_RE = r"[^\-.0-9]"


@dataclass(frozen=True)
class TabularArtefact:
    best_disc_model: str = "best-disc-model"
    mean_best_disc_model: str = "mean-best-disc-model"
    not_best_disc_model: str = "not-best-disc-model"
    last_epoch_model: str = "last-epoch-model"

    @staticmethod
    def artefacts():
        return [field.default for field in fields(TabularArtefact)]


@dataclass(frozen=True)
class ModelFileName:
    rtf_config_json: str = "rtf_config.json"
    rtf_model_pt: str = "rtf_model.pt"

    @staticmethod
    def names():
        return [field.default for field in fields(ModelFileName)]


@dataclass(frozen=True)
class ModelType:
    tabular: str = "tabular"
    relational: str = "relational"

    @staticmethod
    def types():
        return [field.default for field in fields(ModelType)]


@dataclass(frozen=True)
class ColDataType:
    NUMERIC: str = "NUMERIC"
    DATETIME: str = "DATETIME"
    CATEGORICAL: str = "CATEGORICAL"

    @staticmethod
    def types():
        return [field.default for field in fields(ColDataType)]


@dataclass(frozen=True)
class SpecialTokens:
    UNK: str = "[UNK]"
    SEP: str = "[SEP]"
    PAD: str = "[PAD]"
    CLS: str = "[CLS]"
    MASK: str = "[MASK]"
    BOS: str = "[BOS]"
    EOS: str = "[EOS]"
    BMEM: str = "[BMEM]"
    EMEM: str = "[EMEM]"
    RMASK: str = "[RMASK]"
    SPTYPE: str = "[SPTYPE]"

    @staticmethod
    def tokens():
        return [field.default for field in fields(SpecialTokens)]


def get_uuid():
    return uuid.uuid4().hex


def fix_multi_decimal(v):
    if v.count(".") > 1:
        v = v.split(".")
        v = ".".join([v[0], "".join(v[1:])])
    return v


def build_vocab(df: pd.DataFrame = None, special_tokens=None, add_columns: bool = True):
    assert (
        df is not None
    ) or special_tokens, "At least one of `df` or `special_tokens` must not be None."

    if add_columns and df is not None:
        # We limit this feature to data that are likely
        # to have been processed using the convention imposed
        # where we keep track of the column index in the processed data.
        assert df.columns.str[0].str.isdigit().all()

    id2token = {}
    curr_id = 0
    if special_tokens:
        id2token.update(dict(enumerate(special_tokens)))
        curr_id = max(id2token) + 1
    column_token_ids = {}

    if df is not None:
        for col in df.columns:
            id2token.update(dict(enumerate(sorted(df[col].unique()), curr_id)))
            column_token_ids[col] = list(range(curr_id, max(id2token) + 1))
            curr_id = max(id2token) + 1

        if add_columns:
            id2token.update(
                dict(
                    enumerate(
                        [extract_processed_column(col) for col in df.columns], curr_id
                    )
                )
            )

            # Add this here to prevent a semantic error in case we add another
            # set of tokens later.
            curr_id = max(id2token) + 1

    token2id = {v: k for k, v in id2token.items()}

    return dict(
        id2token=id2token,
        token2id=token2id,
        column_token_ids=column_token_ids,
    )


def process_numeric_data(
    series: pd.Series,
    max_len: int = 10,
    numeric_precision: int = 4,
    transform_data: Dict = None,
) -> Tuple[pd.Series, Dict]:
    is_transform = True

    if transform_data is None:
        transform_data = dict()
        is_transform = False

    if is_transform:
        warnings.warn(
            "Default values will be overridden because transform_data was passed..."
        )
        max_len = transform_data["max_len"]
        numeric_precision = transform_data["numeric_precision"]
    else:
        transform_data["max_len"] = max_len
        transform_data["numeric_precision"] = numeric_precision

    # Note that at this point, we should have casted int-like values to
    # pd.Int64Dtype but just to be very sure, let's do that again here.
    try:
        series = series.astype(pd.Int64Dtype())
    except TypeError:
        pass
    except ValueError:
        pass

    if series.dtype == pd.Int64Dtype():
        series = series.astype(str)
    else:
        # We convert float-like values to string with the specified
        # maximum precision.
        # NOTE: Don't use series.round(numeric_precision).astype(str)
        # In some cases, this introduces scientific notation in the
        # string version causing "invalid" values.

        # Note that our purpose for doing this is to actually truncate
        # the precision and not increase the precision.
        # So, we strip the right trailing zeros because the formatting
        # pads the series to the numeric_precision even when not needed.
        series = series.map(lambda x: f"{x:.{numeric_precision}f}").str.rstrip("0")

    # Get the most significant digit
    if is_transform:
        mx_sig = transform_data["mx_sig"]
    else:
        mx_sig = series.str.find(".").max()
        transform_data["mx_sig"] = int(mx_sig)

    if mx_sig <= 0:
        # The data has no decimal point.
        # Pad the data with leading zeros if not
        # aligned to the largest value.
        # We also don't apply the max_len to integral
        # valued data because it will basically
        # remove important information.
        if is_transform:
            zfill = transform_data["zfill"]
        else:
            zfill = series.map(len).max()
            transform_data["zfill"] = int(zfill)
        series = series.str.zfill(zfill)
    else:
        # Make sure that we don't exessively truncate the data.
        # The max_len should be greater than the mx_sig.
        # Add a +1 to generate a minimum of tenth place resolution
        # for this data.
        assert max_len > (
            mx_sig + 1
        ), f"The target length {max_len} of the data doesn't include the numeric precision at {mx_sig}. Increase max_len to at least {max_len + (mx_sig + 2 - max_len)}."

        # Left align first based on the magnitude of the values.
        # We compute the difference in the most significant digits
        # of all values with respect to the largest value.
        # We then pad a leading zero to values with lower most significant
        # digits.
        # For example we have the values 1029.61 and 4.269. This will
        # determine that 1029.61 has the largest magnitude, with most significant
        # digit of 4. It will pad the value 4.269 with three zeros and convert it
        # to 0004.269.
        series = (mx_sig - series.str.find(".")).map(lambda x: "0" * x) + series
        series = series.str[:max_len]

        # We additionally apply left justify to align based on the trailing precision.
        # For example, we have 1029.61 and 0004.269 as values. This time we transform the first
        # value to become 1029.610 to align with the precision of the second value.
        if is_transform:
            ljust = transform_data["ljust"]
        else:
            ljust = series.map(len).max()
            transform_data["ljust"] = int(ljust)

        series = series.str.ljust(ljust, "0")

    # If a number has a negative sign, make sure that it is placed properly.
    series.loc[series.str.contains("-", regex=False)] = "-" + series.loc[
        series.str.contains("-", regex=False)
    ].str.replace("-", "", regex=False)

    return series, transform_data


def process_datetime_data(
    series, transform_data: Dict = None
) -> Tuple[pd.Series, Dict]:
    # Get the max_len from the current time.
    # This will be ignored later if the actual max_len
    # is shorter.
    max_len = len(str(int(time.time())))

    # Convert the datetimes to
    # their equivalent timestamp values.

    # Make sure that we don't convert the NaT
    # to some integer.
    series = series.copy()

    # Track null values (NaT)
    null_idx = series.isnull()

    # Convert to the numerical representation
    # of the datetime (UNIX timestamp)
    series = (series.astype("int64") / 1e9)

    # Fill NA
    series.loc[null_idx] = pd.NA

    # Cast as integer type
    series = series.astype("Int64")

    # Take the mean value to re-align the data.
    # This will help reduce the scale of the numeric
    # data that will need to be generated. Let's just
    # add this offset back later before casting.
    mean_date = None

    if transform_data is None:
        mean_date = int(series.mean())
        series -= mean_date
    else:
        # The mean_date should have been
        # stored during fitting.
        series -= transform_data["mean_date"]

    # Then apply the numeric data processing
    # pipeline.
    series, transform_data = process_numeric_data(
        series,
        max_len=max_len,
        numeric_precision=0,
        transform_data=transform_data,
    )

    # Store the `mean_date` here because `process_numeric_data`
    # expects a None transform_data during fitting.
    if mean_date is not None:
        transform_data["mean_date"] = mean_date

    return series, transform_data


def process_categorical_data(series: pd.Series) -> pd.Series:
    # Simply convert the categorical data to string.
    return series.astype(str)


def encode_partition_numeric_col(col, tr, col_zfill):
    # Generate a derived column name for the partitioned
    # numeric data. For example: Latitude_00, Latitude_01, etc.
    return [f"{col}_{str(i).zfill(col_zfill)}" for i in range(len(tr.columns))]


def decode_partition_numeric_col(partition_col):
    # Decode the encoded column for the partitioned numeric data.
    # For example: Latitude_00 -> Latitude, Latitude_01 -> Latitude, etc.
    return "_".join(partition_col.split("_")[:-1])


def tokenize_numeric_col(series: pd.Series, nparts=2, col_zfill=2):
    # After normalizing the numeric values, we then segment
    # them based on a fixed partition size (nparts).
    col = series.name
    max_len = series.map(len).min()

    # Take the observations that have non-numeric characters.
    # These are NaNs.
    nan_obs = series.str.contains(INVALID_NUMS_RE, regex=True)

    if nparts > max_len > 2:
        # Allow minimum of 0-99 as acceptable singleton range.
        raise ValueError(
            f"Partition size {nparts} is greater than the value length {max_len}. Consider reducing the number of partitions..."
        )
    mx = series.map(len).max()

    tr = pd.concat([series.str[i : i + nparts] for i in range(0, mx, nparts)], axis=1)

    # Replace values with NUMERIC_NA_TOKEN
    tr.loc[nan_obs] = NUMERIC_NA_TOKEN * nparts

    tr.columns = encode_partition_numeric_col(col, tr, col_zfill)

    return tr


def encode_column_values(series):
    return series.name + SPECIAL_COL_SEP + series


def decode_column_values(series):
    # return series.str.split(SPECIAL_COL_SEP).map(lambda x: x[-1])
    return series.map(lambda x: x.split(SPECIAL_COL_SEP)[-1])


def encode_processed_column(idx, dtype, col):
    # The idx indicates the position of the column in the original
    # data.
    # The dtype corresponds to the data type of the column.
    # The col is the actual column name of the data in the table.
    assert dtype in ColDataType.types()
    return f"{idx}{SPECIAL_COL_SEP}{dtype}{SPECIAL_COL_SEP}{col}"


def decode_processed_column(col):
    # Reverts the process in `encode_processed_column`
    # Do nothing if the format is not as expected.

    if not any([dtype in col for dtype in ColDataType.types()]):
        return col

    # Pattern is: [0-9]+___(NUMERIC|DATETIME|CATEGORICAL)___
    return re.sub(
        f"[0-9]+{SPECIAL_COL_SEP}({'|'.join(ColDataType.types())}){SPECIAL_COL_SEP}",
        "",
        col,
    )


def extract_processed_column(col):
    # Extracts the data generated by `encode_processed_column`
    # Return None if not found.

    # Pattern is: ([0-9]+___(NUMERIC|DATETIME|CATEGORICAL)___.*)?____
    match = re.match(
        f"[0-9]+{SPECIAL_COL_SEP}({'|'.join(ColDataType.types())})",
        col,
    )

    return match.group(0) if match else None


def is_numeric_col(col):
    return f"{SPECIAL_COL_SEP}{ColDataType.NUMERIC}{SPECIAL_COL_SEP}" in col


def is_datetime_col(col):
    return f"{SPECIAL_COL_SEP}{ColDataType.DATETIME}{SPECIAL_COL_SEP}" in col


def is_categorical_col(col):
    return f"{SPECIAL_COL_SEP}{ColDataType.CATEGORICAL}{SPECIAL_COL_SEP}" in col


def is_numeric_datetime_col(col):
    return is_numeric_col(col) or is_datetime_col(col)


def process_data(
    df: pd.DataFrame,
    numeric_max_len=10,
    numeric_precision=4,
    numeric_nparts=2,
    first_col_type=None,
    col_transform_data: Dict = None,
    target_col: str = None,
) -> Tuple[pd.DataFrame, Dict]:
    # This should receive a dataframe with dtypes that have already been
    # properly categorized between numeric and categorical.
    # Date type can be converted as UNIX timestamps.
    assert first_col_type in [None, ColDataType.CATEGORICAL, ColDataType.NUMERIC]

    df = df.copy()

    # Unify the variable for missing data
    df = df.fillna(pd.NA)

    # Force cast integral values to Int64Dtype dtype
    # to save precision if they are represented as float.
    for c in df:
        try:
            if pd.api.types.is_datetime64_any_dtype(df[c].dtype):
                # Don't cast datetime types.
                continue

            if pd.api.types.is_numeric_dtype(df[c].dtype):
                # Only cast if the column is explicitly numeric type.
                df[c] = df[c].astype(pd.Int64Dtype())
        except TypeError:
            pass
        except ValueError:
            pass

    if target_col is not None:
        assert (
            first_col_type is None
        ), "Implicit ordering of columns when teacher-forcing of target is used is not supported yet!"
        tf_col_name = f"{TEACHER_FORCING_PRE}_{target_col}"
        assert (
            tf_col_name not in df.columns
        ), f"The column name ({tf_col_name}) must not be in the raw data. Found instead..."

        target_ser = df[target_col].copy()
        target_ser.name = tf_col_name
        df = pd.concat([target_ser, df], axis=1)

    # Rename the columns to encode the original order by adding a suffix of increasing
    # integer values.
    num_cols = len(str(len(df.columns)))
    col_idx = {col: f"{str(i).zfill(num_cols)}" for i, col in enumerate(df.columns)}

    # Create a dataframe that will hold the processed data
    processed_series = []

    # Process numerical data
    numeric_cols = df.select_dtypes(include=np.number).columns

    if col_transform_data is None:
        col_transform_data = dict()

    for c in numeric_cols:
        col_name = encode_processed_column(col_idx[c], ColDataType.NUMERIC, c)
        _col_transform_data = col_transform_data.get(c)
        series, transform_data = process_numeric_data(
            df[c],
            max_len=numeric_max_len,
            numeric_precision=numeric_precision,
            transform_data=_col_transform_data,
        )
        if _col_transform_data is None:
            # This means that no transform data is available
            # before the processing.
            col_transform_data[c] = transform_data
        series.name = col_name
        processed_series.append(series)

    # Process datetime data
    datetime_cols = df.select_dtypes(include="datetime").columns

    for c in datetime_cols:
        col_name = encode_processed_column(col_idx[c], ColDataType.DATETIME, c)

        _col_transform_data = col_transform_data.get(c)
        series, transform_data = process_datetime_data(
            df[c],
            transform_data=_col_transform_data,
        )
        if _col_transform_data is None:
            # This means that no transform data is available
            # before the processing.
            col_transform_data[c] = transform_data
        series.name = col_name
        processed_series.append(series)

    processed_df = pd.concat([pd.DataFrame()] + processed_series, axis=1)

    if not processed_df.empty:
        # Combine the processed numeric and datetime data.
        processed_df = pd.concat(
            [
                tokenize_numeric_col(processed_df[col], nparts=numeric_nparts)
                for col in processed_df.columns
            ],
            axis=1,
        )

    # NOTE: The categorical data should be the last to be processed!
    categorical_cols = df.columns.difference(numeric_cols).difference(datetime_cols)

    if not categorical_cols.empty:
        # Process the rest of the data, assumed to be categorical values.
        processed_df = pd.concat(
            [
                processed_df,
                *(
                    process_categorical_data(df[c]).rename(
                        encode_processed_column(col_idx[c], ColDataType.CATEGORICAL, c)
                    )
                    for c in categorical_cols
                ),
            ],
            axis=1,
        )

    # Get the different sets of column types
    cat_cols = processed_df.columns[
        processed_df.columns.str.contains(ColDataType.CATEGORICAL)
    ]
    numeric_cols = processed_df.columns[
        ~processed_df.columns.str.contains(ColDataType.CATEGORICAL)
    ]

    if first_col_type == ColDataType.CATEGORICAL:
        df = processed_df[cat_cols.union(numeric_cols, sort=False)]
    elif first_col_type == ColDataType.NUMERIC:
        df = processed_df[numeric_cols.union(cat_cols, sort=False)]
    else:
        # Reorder columns to the original order
        df = processed_df[sorted(processed_df.columns)]

    for c in df.columns:
        # Add the column name as part of the value.
        df[c] = encode_column_values(df[c])

    return df, col_transform_data


def get_token_id(
    token: str, vocab_token2id: Dict[str, int], mask_rate: float = 0
) -> int:
    token_id = vocab_token2id.get(token, vocab_token2id[SpecialTokens.UNK])
    if mask_rate > 0:
        token_id = (
            vocab_token2id[SpecialTokens.RMASK]
            if random.random() < mask_rate
            else token_id
        )

    return token_id


def get_input_ids(
    example,
    vocab: Dict,
    columns: List,
    mask_rate: float = 0,
    return_label_ids: Optional[bool] = True,
    return_token_type_ids: Optional[bool] = False,
    affix_bos: Optional[bool] = True,
    affix_eos: Optional[bool] = True,
) -> Dict:
    # Raise an assertion error while the implementation
    # is not yet ready.
    assert return_token_type_ids is False
    input_ids: List[int] = []
    token_type_ids: List[int] = []

    if affix_bos:
        input_ids.append(vocab["token2id"][SpecialTokens.BOS])
        if return_token_type_ids:
            token_type_ids.append(vocab["token2id"][SpecialTokens.SPTYPE])

    for k in columns:
        input_ids.append(get_token_id(example[k], vocab["token2id"], mask_rate))
        if return_token_type_ids:
            col_name = extract_processed_column(k)
            token_type_ids.append(vocab["token2id"][col_name])

    if affix_eos:
        input_ids.append(vocab["token2id"][SpecialTokens.EOS])
        if return_token_type_ids:
            token_type_ids.append(vocab["token2id"][SpecialTokens.SPTYPE])

    data = dict(input_ids=input_ids)

    if return_label_ids:
        data["label_ids"] = input_ids

    if return_token_type_ids:
        data["token_type_ids"] = token_type_ids

    return data


def make_dataset(
    df: pd.DataFrame,
    vocab: Dict,
    mask_rate: float = 0,
    affix_eos: bool = True,
    return_token_type_ids: bool = False,
) -> Dataset:
    # Load the dataframe into a HuggingFace Dataset
    training_dataset = Dataset.from_pandas(df, preserve_index=False)

    # Create the input_ids and label_ids columns
    logging.info("Creating the input_ids and label_ids columns...")

    return training_dataset.map(
        lambda example: get_input_ids(
            example,
            vocab,
            df.columns,
            mask_rate=mask_rate,
            affix_eos=affix_eos,
            return_token_type_ids=return_token_type_ids,
        ),
        remove_columns=training_dataset.column_names,
    )


def get_relational_input_ids(
    example,
    input_idx,
    vocab,
    columns,
    output_dataset,
    in_out_idx,
    output_max_length: Optional[int] = None,
    return_token_type_ids: bool = False,
) -> dict:
    # Start with 2 to take into account the [BOS] and [EOS] tokens
    sequence_len = 2

    # Build the input_ids for the encoder
    input_payload = get_input_ids(
        example,
        vocab["encoder"],
        columns,
        return_label_ids=False,
        return_token_type_ids=return_token_type_ids,
        affix_bos=True,
        affix_eos=True,
    )
    input_ids = input_payload["input_ids"]
    token_type_ids = input_payload.get("token_type_ids")

    # Build the label_ids for the decoder
    output_idx = in_out_idx[input_idx]

    valid = True

    label_ids = [vocab["decoder"]["token2id"][SpecialTokens.BOS]]
    if len(output_idx) > 0:
        for ids in output_dataset.select(output_idx)["input_ids"]:
            # Pad each observation with the [BMEM] and [EMEM] tokens

            tmp_label_ids = [vocab["decoder"]["token2id"][SpecialTokens.BMEM]]
            tmp_label_ids.extend(ids)
            tmp_label_ids.append(vocab["decoder"]["token2id"][SpecialTokens.EMEM])

            if output_max_length:
                if (sequence_len + len(tmp_label_ids)) > output_max_length:
                    # This exceeds the expected limit.
                    # Drop this observation.
                    valid = False
                    break

            label_ids.extend(tmp_label_ids)
            sequence_len += len(tmp_label_ids)

    label_ids.append(vocab["decoder"]["token2id"][SpecialTokens.EOS])

    payload = dict(
        input_ids=input_ids,
        # The variable `labels` is used in the EncoderDecoder model
        # instead of `label_ids`.
        labels=label_ids if valid else None,
    )

    if token_type_ids is not None:
        payload["token_type_ids"] = token_type_ids

    return payload


def make_relational_dataset(
    in_df: pd.DataFrame,
    out_df: pd.DataFrame,
    vocab: dict,
    in_out_idx: dict,
    mask_rate=0,
    output_max_length: Optional[int] = None,
    return_token_type_ids: bool = False,
) -> Dataset:
    # Relational data
    # Load the dataframe into a HuggingFace Dataset
    encoder_dataset = Dataset.from_pandas(in_df, preserve_index=False)

    # Load the dataframe into a HuggingFace Dataset
    decoder_dataset = Dataset.from_pandas(out_df, preserve_index=False)
    # Do not add [BOS] and [EOS] here. This will be handled
    # in the creation of the training_dataset in `get_relational_input_ids`.
    decoder_dataset = decoder_dataset.map(
        lambda example: get_input_ids(
            example,
            vocab["decoder"],
            out_df.columns,
            mask_rate=mask_rate,
            return_label_ids=False,
            return_token_type_ids=return_token_type_ids,
            affix_bos=False,
            affix_eos=False,
        ),
        remove_columns=decoder_dataset.column_names,
    )

    training_dataset = encoder_dataset.map(
        lambda example, idx: get_relational_input_ids(
            example,
            idx,
            vocab,
            in_df.columns,
            decoder_dataset,
            in_out_idx,
            output_max_length,
        ),
        remove_columns=encoder_dataset.column_names,
        with_indices=True,
    )

    # If the output_max_length variable is specified, filter
    # observations that exceed this length. The
    # `get_relational_input_ids` should have set the
    # `labels` to None if the output exceeds `output_max_length`.
    if output_max_length:
        init_data_length = training_dataset.shape[0]

        training_dataset = training_dataset.filter(
            lambda example: example["labels"] is not None
        )

        removed_count = init_data_length - training_dataset.shape[0]
        if removed_count > 0:
            warnings.warn(
                f"A total of {removed_count} out of {init_data_length} has been removed from the training data because they exceeded the `output_max_length` of {output_max_length}."
            )

    return training_dataset



from typing import Any, Dict, Optional, Tuple, Union

import numpy as np
import pandas as pd
from shapely.geometry import MultiPolygon, Point, Polygon


class ValidatorBase:
    def __init__(self) -> None:
        pass

    def validate(self, *args: Any, **kwargs: Any) -> bool:
        raise NotImplementedError


class RangeValidator(ValidatorBase):
    def __init__(
        self,
        min_val: Union[float, int, np.number],
        max_val: Union[float, int, np.number],
    ) -> None:
        super().__init__()

        self.min_val = min_val
        self.max_val = max_val

    def validate(  # type: ignore
        self, val: Union[float, int, np.number], *args: Any, **kwargs: Any
    ) -> bool:
        return self.min_val <= val <= self.max_val


class GeoValidator(ValidatorBase):
    def __init__(self, geo_bound: Union[Polygon, MultiPolygon]) -> None:
        super().__init__()

        self.geo_bound = geo_bound

    def validate(self, lon: float, lat: float) -> bool:  # type: ignore
        p = Point(lon, lat)

        return self.geo_bound.contains(p)


class ObservationValidator(ValidatorBase):
    def __init__(
        self, validators: Optional[Dict[str, Tuple[ValidatorBase, Tuple[str]]]] = None
    ) -> None:
        super().__init__()
        self.validators = validators or {}

    def validate(self, series: pd.Series) -> bool:  # type: ignore
        is_valid = True
        for vname in self.validators:
            validator, cols = self.validators[vname]
            is_valid = is_valid and validator.validate(*(series[c] for c in cols))

            if not is_valid:
                break

        return is_valid

    def validate_df(self, df: pd.DataFrame) -> pd.Series:
        return df.apply(self.validate, axis=1)

    def add_validator(self, name, validator, cols) -> None:
        self.validators[name] = (validator, cols)

    def remove_validator(self, name) -> Tuple:
        return self.validators.pop(name)



from dataclasses import dataclass
from typing import Optional

import numpy as np
import torch


@dataclass
class RelationalDataCollator:
    """
    Data collator that will dynamically pad the inputs received, as well as the labels.
    Adopted from the DataCollatorForSeq2Seq:
     https://github.com/huggingface/transformers/blob/v4.24.0/src/transformers/data/data_collator.py#L510

    Args:
        max_length (`int`, *optional*):
            Maximum length of the returned list and optionally padding length (see above).
        pad_to_multiple_of (`int`, *optional*):
            If set will pad the sequence to a multiple of the provided value.
            This is especially useful to enable the use of Tensor Cores on NVIDIA hardware with compute capability >=
            7.5 (Volta).
        label_pad_token_id (`int`, *optional*, defaults to -100):
            The id to use when padding the labels (-100 will be automatically ignored by PyTorch loss functions).
        return_tensors (`str`):
            The type of Tensor to return. Allowable values are "np", "pt" and "tf".
    """

    max_length: Optional[int] = None
    pad_to_multiple_of: Optional[int] = None
    label_pad_token_id: int = -100
    return_tensors: str = "pt"

    def __call__(self, features, return_tensors=None):
        if return_tensors is None:
            return_tensors = self.return_tensors

        labels = (
            [feature["labels"] for feature in features]
            if "labels" in features[0].keys()
            else None
        )

        # We have to pad the labels before calling `tokenizer.pad` as this method won't pad them and needs them of the
        # same length to return tensors.
        if labels is not None:
            max_label_length = max(len(label) for label in labels)
            if self.pad_to_multiple_of is not None:
                max_label_length = (
                    (max_label_length + self.pad_to_multiple_of - 1)
                    // self.pad_to_multiple_of
                    * self.pad_to_multiple_of
                )

            for feature in features:
                remainder = [self.label_pad_token_id] * (
                    max_label_length - len(feature["labels"])
                )
                if isinstance(feature["labels"], list):
                    feature["labels"] = feature["labels"] + remainder
                else:
                    # Pad always at the right.
                    feature["labels"] = np.concatenate(
                        [feature["labels"], remainder]
                    ).astype(np.int64)

        labels = [feature["labels"] for feature in features]
        input_ids = [feature["input_ids"] for feature in features]

        if return_tensors == "np":
            labels = np.vstack(labels)
            input_ids = np.vstack(input_ids)
        elif return_tensors == "pt":
            labels = torch.vstack([torch.tensor(label) for label in labels])
            input_ids = torch.vstack([torch.tensor(ii) for ii in input_ids])
        elif return_tensors == "tf":
            raise ValueError("Tensorflow tensor is not supported yet.")

        return dict(
            labels=labels,
            input_ids=input_ids,
        )



import math
import os
import random
from typing import Any, List, Optional, Union

import joblib
import numpy as np
import pandas as pd
import sklearn
from scipy import stats
from sklearn.compose import ColumnTransformer
from sklearn.decomposition import TruncatedSVD
from sklearn.metrics.pairwise import manhattan_distances
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import OneHotEncoder, StandardScaler
from tqdm.auto import tqdm


class SyntheticDataBench:
    """This class handles all the assessments
    needed for testing the synthetic data.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        target_col: str,
        categorical: bool,
        target_pos_val: Any = None,
        test_size: float = 0.2,
        test_df: Optional[pd.DataFrame] = None,
        random_state: int = 1029,
    ) -> None:
        assert (
            test_size < 1
        ), "The test_size must be a fraction of the data, and should be less than 1."
        self.random_state = random_state

        # Target column in the data for the ML efficiency
        # measure.
        self.target_col = target_col
        self.categorical = categorical
        self.target_pos_val = target_pos_val

        if test_df is not None:
            self.test_df = test_df
            self.train_df = data
            self.test_size = None
        else:
            self.test_df = data.sample(
                frac=test_size, replace=False, random_state=self.random_state
            )
            self.train_df = data.loc[data.index.difference(self.test_df.index)]
            self.test_size = test_size

        self.train_df = pd.concat(
            [self.train_df.drop(target_col, axis=1), self.train_df[target_col]], axis=1
        )
        self.test_df = pd.concat(
            [self.test_df.drop(target_col, axis=1), self.test_df[target_col]], axis=1
        )

        self.n_test: int = len(self.test_df)
        self.n_train: int = len(self.train_df)
        self.synth_train_df: pd.DataFrame = None
        self.synth_test_df: pd.DataFrame = None

    def register_synthetic_data(self, synthetic: pd.DataFrame):
        """
        Registers synthetic data for the assessment.

        The synthetic data is split into training and test sets according
        to the values of n_train and n_test. The split is done by sampling
        the data without replacement.

        Args:
        synthetic: A DataFrame containing synthetic data.
        The DataFrame must have at least as many rows as n_train + n_test.

        Returns:
        None
        """

        assert synthetic.shape[0] >= (self.n_test + self.n_train)
        self.synth_train_df = synthetic.sample(
            n=self.n_train, replace=False, random_state=self.random_state
        )
        self.synth_test_df = synthetic.loc[
            synthetic.index.difference(self.synth_train_df.index)
        ].sample(n=self.n_test, replace=False, random_state=self.random_state)

    @staticmethod
    def compute_distance_to_closest_records(
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        n_test: int,
        distance: manhattan_distances = manhattan_distances,
    ) -> pd.Series:
        """
        original: The dataframe of the training data used to train the generative model.
        synthetic: The dataframe generated by the generative model or any data we want to compare
         with the original data.
        n_test: The number of observations we want to compare with the original data
         from the synthetic data. Ideally, this should be the same size as the test data.
        """
        assert n_test <= len(synthetic)
        synthetic = synthetic.iloc[:n_test]

        distances: np.ndarray = distance(original, synthetic)
        return pd.Series(distances.min(axis=0), index=synthetic.index)

    @staticmethod
    def measure_ml_efficiency(
        model: sklearn.base.BaseEstimator,
        train: pd.DataFrame,
        synthetic: pd.DataFrame,
        test: pd.DataFrame,
        target_col: str,
        random_state: int = 1029,
    ) -> pd.DataFrame:
        """
        This function trains the provided model on the original and synthetic training data,
        and then uses the trained models to make predictions on the test data. It returns a
        dataframe containing the actual values and predictions from both training sets. This
        dataframe can be used to compare the performance of the model trained on the original
        data with the model trained on the synthetic data.

        Parameters:
        model (sklearn.base.BaseEstimator): The model to be trained and used for prediction.
        train (pd.DataFrame): The original training data.
        synthetic (pd.DataFrame): The synthetic training data generated by a generative model.
         Must have the same size as the `train`.
        test (pd.DataFrame): The test data to be used for prediction.
        target_col (str): The name of the target column in the train and test data.

        Returns:
        pd.DataFrame: A dataframe containing the actual values and predictions from both
         training sets.
        """

        random.seed(random_state)
        np.random.seed(random_state)

        assert train.shape[0] == synthetic.shape[0]

        # Train the model on the original training data
        model.fit(train.drop(target_col, axis=1), train[target_col])

        # Make predictions on the test data using the original training data
        try:
            original_predictions = model.predict_proba(test.drop(target_col, axis=1))
        except AttributeError:
            original_predictions = model.predict(test.drop(target_col, axis=1))

        # Train the model on the synthetic training data
        model.fit(synthetic.drop(target_col, axis=1), synthetic[target_col])

        # Make predictions on the test data using the synthetic training data
        try:
            synthetic_predictions = model.predict_proba(test.drop(target_col, axis=1))
        except AttributeError:
            synthetic_predictions = model.predict(test.drop(target_col, axis=1))

        # Return a dataframe with the actual values and predictions from both training sets
        return pd.DataFrame(
            {
                "actual": test[target_col],
                "original_predictions": original_predictions,
                "synthetic_predictions": synthetic_predictions,
            }
        )

    @staticmethod
    def preprocess_data(
        data: pd.DataFrame,
        other: Union[pd.DataFrame, List[pd.DataFrame]] = None,
        fillna: bool = True,
    ) -> dict:
        """Preprocesses a DataFrame containing mixed data types and returns a feature matrix.

        The function first extracts the categorical and numerical columns from the DataFrame,
        and then applies a processing pipeline that one-hot encodes the categorical features
        and standardizes the numerical features.

        Args:
            data (pandas.DataFrame): A DataFrame containing mixed data types.

        Returns:
            dict:
             - preprocessor: The trained feature processor pipeline.
             - column_names: The new column names for the processed data.
             - data: A feature matrix containing only numerical values for the input data.
             - other (optional): A feature matrix containing only numerical values
              for the input other.
        """
        # Define a processing pipeline for the data
        index = data.index
        numeric_features = data.select_dtypes(include="number").columns
        categorical_features = data.select_dtypes(include="object").columns

        transformers = []
        column_names = []
        numeric_cols = []
        categorical_cols = []

        if not numeric_features.empty:
            numeric_transformer = StandardScaler()
            transformers.append(("num", numeric_transformer, numeric_features))
            column_names.extend(numeric_features)
            numeric_cols.extend(numeric_features)

        if not categorical_features.empty:
            categorical_transformer = OneHotEncoder(handle_unknown="ignore")
            transformers.append(("cat", categorical_transformer, categorical_features))

        preprocessor = ColumnTransformer(transformers=transformers)
        preprocessor.fit(data)

        if not categorical_features.empty:
            for transf in preprocessor.transformers_:
                if isinstance(transf[1], OneHotEncoder):
                    column_names.extend(transf[1].get_feature_names_out())
                    categorical_cols.extend(transf[1].get_feature_names_out())

        data = preprocessor.transform(data)

        payload = dict(
            preprocessor=preprocessor,
            numeric_cols=numeric_cols,
            column_names=column_names,
            categorical_cols=categorical_cols,
            data=pd.DataFrame(
                # Sometimes the transform method returns a sparse matrix.
                # So we convert it to an array if the data is not a numpy
                # array.
                data if isinstance(data, np.ndarray) else data.toarray(),
                columns=column_names,
                index=index,
            ),
        )
        if fillna:
            payload["data"] = payload["data"].fillna(payload["data"].mean())

        if other is not None:
            transformed_other = []
            is_df = False

            if isinstance(other, pd.DataFrame):
                is_df = True
                other = [other]

            for _other in other:
                index = _other.index
                _other = preprocessor.transform(_other)

                _other = pd.DataFrame(
                    _other if isinstance(_other, np.ndarray) else _other.toarray(),
                    columns=column_names,
                    index=index,
                )
                if fillna:
                    _other = _other.fillna(_other.mean())
                transformed_other.append(_other)

            payload["other"] = transformed_other[0] if is_df else transformed_other

        return payload

    @staticmethod
    def compute_discriminator_predictions(
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        test: pd.DataFrame,
        model: sklearn.base.BaseEstimator,
        random_state: int = 1029,
    ) -> dict:
        """
        Builds a discriminator model that attempts to distinguish between original and synthetic data.

        The function first preprocesses the data by extracting the categorical and numerical columns,
        then applies a processing pipeline that one-hot encodes the categorical features
        and standardizes the numerical features.
        Next, it adds labels to the original and synthetic data to indicate which is which,
        then combines the data into one DataFrame and splits it into training and test sets.
        Finally, it trains a classifier model on the training data and returns the model.

        Args:
            original (pandas.DataFrame): A DataFrame containing original data.
            synthetic (pandas.DataFrame): A DataFrame containing synthetic data.
            model (Type[LogisticRegression]): A type of scikit-learn model to use.
                Defaults to LogisticRegression.
            test_size (float): The proportion of data to include in the test set.
                Defaults to 0.2.
            random_state (int): The random seed to use for splitting the data.
                Defaults to 1029.

        Returns:
            dict:
             - y_test: Labels for the test/synthetic test data.
             - y_preds: Predictions for the label.
        """
        assert synthetic.shape[0] >= (len(original) + len(test))

        train_synthetic = synthetic.sample(
            n=len(original), replace=False, random_state=random_state
        )
        test_synthetic = synthetic.loc[
            synthetic.index.difference(train_synthetic.index)
        ].sample(n=len(test), replace=False, random_state=random_state)

        # Preprocess the original and synthetic data
        processed = SyntheticDataBench.preprocess_data(
            data=original, other=[train_synthetic, test, test_synthetic]
        )
        original = processed["data"]
        train_synthetic = processed["other"][0]
        test = processed["other"][1]
        test_synthetic = processed["other"][2]

        label_col_name = "discriminator_label"

        assert label_col_name not in processed["column_names"]

        # Add labels to the original and synthetic data
        original[label_col_name] = 1
        train_synthetic[label_col_name] = 0

        # Combine the original and synthetic data
        combined_data = pd.concat([original, train_synthetic])

        feat_train = combined_data.drop(label_col_name, axis=1)
        y_train = combined_data[label_col_name]
        feat_train = feat_train.fillna(feat_train.mean())

        # Train a logistic regression model on the training data
        model.fit(feat_train, y_train)

        oob_score = None
        try:
            oob_score = model.oob_score_
            print("Discriminator OOB Score:", oob_score)
        except AttributeError:
            pass

        # Transform the test data and the synthetic test data
        test[label_col_name] = 1
        test_synthetic[label_col_name] = 0

        combined_test = pd.concat([test, test_synthetic])
        feat_test = combined_test.drop(label_col_name, axis=1)
        y_test = combined_test[label_col_name]
        feat_test = feat_test.fillna(feat_test.mean())

        # Estimate the discriminator on the test data
        y_preds = model.predict(feat_test)

        return dict(y_test=y_test, y_preds=y_preds, oob_score=oob_score)

    def get_dcr(
        self, is_test: bool = False, distance: manhattan_distances = manhattan_distances
    ) -> pd.Series:
        """Get the DCR values for this experiment."""
        train = self.train_df
        other = self.test_df if is_test else self.synth_train_df

        if (train.dtypes == "object").any():
            proc = self.preprocess_data(data=train, other=other)
            train = proc["data"]
            other = proc["other"]

        return self.compute_distance_to_closest_records(
            train,
            other,
            self.n_test,
            distance=distance,
        )

    def get_ml_efficiency(
        self, model: sklearn.base.BaseEstimator, synthetic: pd.DataFrame = None
    ) -> pd.DataFrame:
        """Get the ML efficiency for this experiment."""
        train = self.train_df.copy()
        test = self.test_df.copy()

        if synthetic is None:
            synthetic = self.synth_train_df

        train_target = train[self.target_col]
        test_target = test[self.target_col]
        synthetic_target = synthetic[self.target_col]

        if self.categorical:
            train_target = 1 * (self.target_pos_val == train_target)
            test_target = 1 * (self.target_pos_val == test_target)
            synthetic_target = 1 * (self.target_pos_val == synthetic_target)

        train = train.drop(self.target_col, axis=1)
        test = test.drop(self.target_col, axis=1)
        synthetic = synthetic.drop(self.target_col, axis=1)

        proc = self.preprocess_data(data=train, other=[test, synthetic])
        train = proc["data"]
        test = proc["other"][0]
        synthetic = proc["other"][1]

        train = pd.concat([train, train_target], axis=1)
        test = pd.concat([test, test_target], axis=1)
        synthetic = pd.concat([synthetic, synthetic_target], axis=1)

        return self.measure_ml_efficiency(
            model=model,
            train=train,
            synthetic=synthetic,
            test=test,
            target_col=self.target_col,
            random_state=self.random_state,
        )

    def get_discriminator_performance(self, model: sklearn.base.BaseEstimator):
        """Compute the discriminator performance for this experiment."""
        return self.compute_discriminator_predictions(
            original=self.train_df,
            synthetic=pd.concat([self.synth_train_df, self.synth_test_df]),
            test=self.test_df,
            model=model,
            random_state=self.random_state,
        )

    @staticmethod
    def compute_data_copying_predictions(
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        test: pd.DataFrame,
        model: sklearn.base.BaseEstimator,
        random_state: int = 1029,
    ) -> dict:
        """
        Builds a discriminator model that attempts to distinguish between original and synthetic data.

        The function first preprocesses the data by extracting the categorical and numerical columns,
        then applies a processing pipeline that one-hot encodes the categorical features
        and standardizes the numerical features.
        Next, it adds labels to the original and synthetic data to indicate which is which,
        then combines the data into one DataFrame and splits it into training and test sets.
        Finally, it trains a classifier model on the training data and returns the model.

        Args:
            original (pandas.DataFrame): A DataFrame containing original data.
            synthetic (pandas.DataFrame): A DataFrame containing synthetic data.
            model (Type[LogisticRegression]): A type of scikit-learn model to use.
                Defaults to LogisticRegression.
            test_size (float): The proportion of data to include in the test set.
                Defaults to 0.2.
            random_state (int): The random seed to use for splitting the data.
                Defaults to 1029.

        Returns:
            dict:
             - y_test: Labels for the test/synthetic test data.
             - y_preds: Predictions for the label.
        """
        # assert synthetic.shape[0] >= (len(original) + len(test))

        # Preprocess the original and test data
        processed = SyntheticDataBench.preprocess_data(original, test)
        preprocessor = processed["preprocessor"]
        column_names = processed["column_names"]
        original = processed["data"]
        test = processed["other"]

        label_col_name = "discriminator_label"

        assert label_col_name not in processed["column_names"]

        # Add labels to the original and test data
        original[label_col_name] = 1
        test[label_col_name] = 0

        # Combine the original and test data
        combined_data = pd.concat([original, test])

        X_train = combined_data.drop(label_col_name, axis=1)
        y_train = combined_data[label_col_name]
        X_train = X_train.fillna(X_train.mean())

        # Train a logistic regression model on the training data
        model.fit(X_train, y_train)

        try:
            print("Discriminator OOB Score:", model.oob_score_)
        except AttributeError:
            pass

        # Transform the test data and the synthetic test data
        synthetic = preprocessor.transform(synthetic)
        synthetic = pd.DataFrame(
            synthetic if isinstance(synthetic, np.ndarray) else synthetic.toarray(),
            columns=column_names,
        )
        X_test = synthetic
        X_test = X_test.fillna(X_test.mean())

        # Estimate the discriminator on the test data
        y_preds = model.predict(X_test)

        return dict(y_preds=y_preds)

    @staticmethod
    def compute_sensitivity_metric(
        original: pd.DataFrame,
        synthetic: pd.DataFrame,
        test: pd.DataFrame,
        qt_max: float = 0.05,
        qt_interval: int = 1000,
        distance: manhattan_distances = manhattan_distances,
        tsvd: TruncatedSVD = None,
        max_col_nums: int = 50,
        use_ks: bool = False,
        verbose: bool = False,
    ) -> float:
        object_dtypes = original.select_dtypes(exclude="number")

        if not object_dtypes.empty or original.shape[1] >= max_col_nums:
            if verbose:
                print("Transforming data with non-numeric values...")
            processed = SyntheticDataBench.preprocess_data(
                original, other=[test, synthetic]
            )

            original = processed["data"]
            test = processed["other"][0]
            synthetic = processed["other"][1]

        if tsvd is None and original.shape[1] >= max_col_nums:
            # We use a truncated SVD with components at least 5 or
            # equivalent to the square root of the original number
            # of variables.
            tsvd = TruncatedSVD(
                n_components=max(5, math.ceil(original.shape[1] ** 0.5))
            )

        if tsvd is not None:
            if verbose:
                print("Applying TruncatedSVD")
            # We reduce the dimensionality of the transformed data
            original = tsvd.fit_transform(original)
            test = tsvd.transform(test)
            synthetic = tsvd.transform(synthetic)

        # We compute the relative distances of each sub data with respect
        # to the original data.
        test_distances: np.ndarray = distance(original, test)
        synth_distances: np.ndarray = distance(original, synthetic)

        # We take the distance of the closest point from each sub data
        # to observations in the original data.
        test_min = test_distances.min(axis=1)
        synth_min = synth_distances.min(axis=1)

        tr_test_min = test_distances.min(axis=0)
        tr_synth_min = synth_distances.min(axis=0)

        # We don't include observations that have duplicates in both
        # the synthetic and the test data.
        fltr = (synth_min == 0) & (test_min == 0)
        test_min = test_min[~fltr]
        synth_min = synth_min[~fltr]

        test_min = np.concatenate([test_min, tr_test_min])
        synth_min = np.concatenate([synth_min, tr_synth_min])

        statistic = None

        if use_ks:
            test_min = test_min[test_min <= np.quantile(test_min, qt_max)]
            synth_min = synth_min[synth_min <= np.quantile(synth_min, qt_max)]

            ks_stat = stats.ks_2samp(test_min, synth_min)

            statistic = ks_stat.statistic

        elif qt_interval <= 1:
            statistic = (synth_min <= np.quantile(test_min, qt_max)).mean()
        else:
            # We define the quantile set to assess the systematic
            # bias in the distance of the synthetic data, if any.
            quantiles = np.linspace(0, qt_max, qt_interval)

            # # The vectorized form is equivalent to the expanded form below:
            # # We do not use the absolute value of the difference so that the
            # # asymptotic value should be closer to 0. Anything that is significantly
            # # different from 0 is anomalous.
            # np.mean([((synth_min <= np.quantile(test_min, qt)).mean() - qt) for qt in np.linspace(0, qt_max, qt_interval)])

            # For each quantile of distances from the test data, we take the proportion
            # of the synthetic data with distance values lower than the value in the given quantile.
            # We expect that the more the model becomes overfitted, the closer to zero the distances
            # coming from the synthetic data is. However, if the data comes from the same distribution,
            # we expect to see this statistic to be closer to zero.
            # We use `<=` so that we can still capture the statistic correctly even when
            # the quantile value is zero.
            statistic = np.mean(
                (
                    synth_min.reshape(1, -1)
                    <= np.quantile(test_min, quantiles).reshape(-1, 1)
                ).mean(axis=1)
                - quantiles
            )

        return statistic

    @staticmethod
    def compute_sensitivity_threshold(
        train_data: pd.DataFrame,
        num_bootstrap: int = 100,
        test_size: int = None,
        frac: float = None,
        qt_max: float = 0.05,
        qt_interval: int = 1000,
        distance: manhattan_distances = manhattan_distances,
        tsvd: TruncatedSVD = None,
        return_values: bool = False,
        quantile: float = 0.95,
        max_col_nums: int = 50,
        use_ks: bool = False,
        full_sensitivity: bool = True,
        sensitivity_orig_frac_multiple: int = 3,
    ) -> Union[float, List]:
        """This method implements a bootstrapped estimation of the
        sensitivity values derived from the training data.

        We compute the sensitivity value for `num_bootstrap` rounds of random split
        of the training data.

        Args:
            quantile: Returns the sensitivity value at the given quantile from
             the bootstrap set. Note that we use quantile > 0.5 because we want to
             detect whether the synthetic data tends to be closer to the training data
             than expected. The statistic computes synth_min < test_min, so if the
             synthetic data systematically copies observation from the training data,
             we expect that the statictic tends to become larger >> 0.
            return_values: Instead of returning a single value based on the `quantile`
             argument, return the full set of boostrap values.
            sensitivity_orig_frac_multiple: The size of the training data relative to the chosen `frac` that will be
             used in computing the sensitivity. The larger this value is, the more robust the sensitivity threshold
             will be. However, `(sensitivity_orig_frac_multiple + 2)` multiplied by `frac` must be less than 1.
        """
        if test_size is not None:
            assert (
                test_size > 1
            ), "The test_size argument corresponds to the number of test samples"
            frac = test_size / len(train_data)

        if frac is None:
            raise ValueError("Either the test_size or frac must be provided")

        assert (
            2 * frac
        ) < 1, "This exceeds the test size and no training data will remain."

        values = []

        if not full_sensitivity:
            # We use the full fraction
            frac = 2 * frac

            # We will be sampling data from the dataset that is
            # `(sensitivity_orig_frac_multiple + 2) * frac` without
            # replacement. We should make sure that we don't exceed this
            # threshold.
            assert (sensitivity_orig_frac_multiple + 2) * frac <= 1

        def bootstrap_inner_loop():
            original: pd.DataFrame = None
            test: pd.DataFrame = None

            if full_sensitivity:
                original, test = train_test_split(train_data, test_size=2 * frac)
                synthetic = test.iloc[: len(test) // 2]
                test = test.loc[test.index.difference(synthetic.index)]
            else:
                source: pd.DataFrame = None
                n_size = int(len(train_data) * frac)
                n_train_size = sensitivity_orig_frac_multiple * n_size
                test_size = n_train_size + (2 * n_size)

                _, source = train_test_split(train_data, test_size=test_size)
                original = source.iloc[:n_train_size]
                synthetic = source.iloc[n_train_size : n_train_size + n_size]
                test = source.iloc[n_train_size + n_size :]

                assert synthetic.shape[0] == test.shape[0]

            return SyntheticDataBench.compute_sensitivity_metric(
                original=original,
                synthetic=synthetic,
                test=test,
                qt_max=qt_max,
                qt_interval=qt_interval,
                distance=distance,
                tsvd=tsvd,
                max_col_nums=max_col_nums,
                use_ks=use_ks,
            )

        n_jobs = 1
        cpu_count = os.cpu_count()

        if cpu_count and cpu_count >= 4:
            n_jobs = min(max(2, cpu_count // 4), 16)

        if n_jobs == 1:
            for _ in tqdm(range(num_bootstrap), desc="Bootstrap round"):
                values.append(bootstrap_inner_loop())
        else:
            print("Using parallel computation!!!")
            with joblib.Parallel(n_jobs=n_jobs) as parallel:
                values = parallel(
                    joblib.delayed(bootstrap_inner_loop)()
                    for _ in tqdm(range(num_bootstrap), desc="Bootstrap round")
                )

        print("Sensitivity threshold summary:")
        print(pd.Series(values).describe())

        return values if return_values else np.quantile(values, quantile)


class SyntheticDataExperiment:
    """
    For each data and model:
    1. Split train/test data -> save data
    2. Train model with train data -> save model
    3. Generate N x train+test synthetic data -> save samples
    4. Perform analysis on the generated data.
    """

    def __init__(
        self,
        data_id: str,
        model_type: str,
        categorical: bool,
        target_col: str,
        target_pos_val: Any = None,
    ) -> None:
        pass



import math
from typing import Callable, Dict, List, Optional, Tuple, Union

import torch
from datasets import Dataset
from torch import nn
from transformers import (
    DataCollator,
    EvalPrediction,
    PreTrainedModel,
    PreTrainedTokenizerBase,
    Trainer,
    TrainerCallback,
    TrainerControl,
    TrainerState,
    TrainingArguments,
    logging,
)
from transformers.optimization import get_scheduler

logger = logging.get_logger(__name__)


class SaveEpochEndCallback(TrainerCallback):
    """This callback forces a checkpoint save at each epoch end."""

    def __init__(self, save_epochs: int = None) -> None:
        super().__init__()

        self.save_epochs = save_epochs

    def on_epoch_end(
        self,
        args: TrainingArguments,
        state: TrainerState,
        control: TrainerControl,
        **kwargs,
    ):
        if self.save_epochs is not None:
            control.should_save = math.ceil(state.epoch) % self.save_epochs == 0
        else:
            control.should_save = True

        return control


class ResumableTrainer(Trainer):
    """This trainer makes the scheduler consistent over pauses
    in the training. The scheduler should return values similar
    to when a training is done either intermittently or continuously
    over the `target_epochs`.
    """

    def __init__(
        self,
        target_epochs: int = None,
        save_epochs: int = None,
        model: Union[PreTrainedModel, nn.Module] = None,
        args: TrainingArguments = None,
        data_collator: Optional[DataCollator] = None,
        train_dataset: Optional[Dataset] = None,
        eval_dataset: Optional[Dataset] = None,
        tokenizer: Optional[PreTrainedTokenizerBase] = None,
        model_init: Callable[[], PreTrainedModel] = None,
        compute_metrics: Optional[Callable[[EvalPrediction], Dict]] = None,
        callbacks: Optional[List[TrainerCallback]] = None,
        optimizers: Tuple[torch.optim.Optimizer, torch.optim.lr_scheduler.LambdaLR] = (
            None,
            None,
        ),
        preprocess_logits_for_metrics: Callable[
            [torch.Tensor, torch.Tensor], torch.Tensor
        ] = None,
    ):
        # Declare here for typing
        self.lr_scheduler: torch.optim.lr_scheduler.LambdaLR = None

        if callbacks is None:
            callbacks = []

        callbacks.append(SaveEpochEndCallback(save_epochs=save_epochs))

        super().__init__(
            model,
            args,
            data_collator,
            train_dataset,
            eval_dataset,
            tokenizer,
            model_init,
            compute_metrics,
            callbacks,
            optimizers,
            preprocess_logits_for_metrics,
        )
        self.target_epochs = target_epochs

    def create_scheduler(
        self, num_training_steps: int, optimizer: torch.optim.Optimizer = None
    ) -> torch.optim.lr_scheduler.LambdaLR:
        """
        Setup the scheduler. The optimizer of the trainer must have been set up either before this method is called or
        passed as an argument.
        Args:
            num_training_steps (int): The number of training steps to do.
        """

        if self.lr_scheduler is None:
            if self.target_epochs is not None:
                # Compute the max_steps based from the
                # `target_epochs`.
                train_dataloader = self.get_train_dataloader()
                len_dataloader = len(train_dataloader)
                num_update_steps_per_epoch = (
                    len_dataloader // self.args.gradient_accumulation_steps
                )
                num_update_steps_per_epoch = max(num_update_steps_per_epoch, 1)

                max_steps = math.ceil(self.target_epochs * num_update_steps_per_epoch)
                num_training_steps = max_steps

            self.lr_scheduler = get_scheduler(
                self.args.lr_scheduler_type,
                optimizer=self.optimizer if optimizer is None else optimizer,
                num_warmup_steps=self.args.get_warmup_steps(num_training_steps),
                num_training_steps=num_training_steps,
            )

        return self.lr_scheduler
